@@ -36,32 +36,33 @@ static const AP4_Track::Type TIDC[adaptive::AdaptiveTree::STREAM_TYPE_COUNT] = {
 
 CSession::CSession(const PROPERTIES::KodiProperties& kodiProps,
                    const std::string& manifestUrl,
-                   const std::map<std::string, std::string>& mediaHeaders,
                    const std::string& profilePath)
   : m_kodiProps(kodiProps),
     m_manifestUrl(manifestUrl),
-    m_mediaHeaders(mediaHeaders),
     m_KodiHost(std::make_unique<CKodiHost>()),
     m_reprChooser(CHOOSER::CreateRepresentationChooser(kodiProps))
 {
   m_KodiHost->SetProfilePath(profilePath);
+  m_KodiHost->SetDebugSaveLicense(kodi::addon::GetSettingBoolean("debug.save.license"));
 
   switch (kodiProps.m_manifestType)
   {
     case PROPERTIES::ManifestType::MPD:
-      m_adaptiveTree = new adaptive::DASHTree(kodiProps, m_reprChooser);
+      m_adaptiveTree = new adaptive::DASHTree(m_reprChooser);
       break;
     case PROPERTIES::ManifestType::ISM:
-      m_adaptiveTree = new adaptive::SmoothTree(kodiProps, m_reprChooser);
+      m_adaptiveTree = new adaptive::SmoothTree(m_reprChooser);
       break;
     case PROPERTIES::ManifestType::HLS:
-      m_adaptiveTree = new adaptive::HLSTree(kodiProps, m_reprChooser);
+      m_adaptiveTree = new adaptive::HLSTree(m_reprChooser);
       break;
     default:
       LOG::LogF(LOGFATAL, "Manifest type not handled");
       return;
   };
 
+  m_adaptiveTree->Configure(kodiProps);
+
   m_settingNoSecureDecoder = kodi::addon::GetSettingBoolean("NOSECUREDECODER");
   LOG::Log(LOGDEBUG, "Setting NOSECUREDECODER value: %d", m_settingNoSecureDecoder);
 
@@ -140,6 +141,19 @@ void CSession::SetSupportedDecrypterURN(std::string& key_system)
       m_dllHelper = std::make_unique<kodi::tools::CDllHelper>();
       if (m_dllHelper->LoadDll(item.Path()))
       {
+#if defined(__linux__) && defined(__aarch64__) && !defined(ANDROID)
+        // On linux arm64, libwidevinecdm.so depends on two dynamic symbols:
+        //   __aarch64_ldadd4_acq_rel
+        //   __aarch64_swp4_acq_rel
+        // These are defined in libssd_wv.so, but to make them available in the main binary's PLT,
+        // we need RTLD_GLOBAL. LoadDll() above uses RTLD_LOCAL, so we use RTLD_NOLOAD here to
+        // switch the flags from LOCAL to GLOBAL.
+        void *hdl = dlopen(item.Path().c_str(), RTLD_NOLOAD | RTLD_GLOBAL | RTLD_LAZY);
+        if (!hdl)
+        {
+          LOG::Log(LOGERROR, "Failed to reload dll in global mode: %s", dlerror());
+        }
+#endif
         CreateDecryptorInstanceFunc startup;
         if (m_dllHelper->RegisterSymbol(startup, "CreateDecryptorInstance"))
         {
@@ -217,12 +231,12 @@ bool CSession::Initialize()
   // Get URN's wich are supported by this addon
   if (!m_kodiProps.m_licenseType.empty())
   {
-    SetSupportedDecrypterURN(m_adaptiveTree->supportedKeySystem_);
-    LOG::Log(LOGDEBUG, "Supported URN: %s", m_adaptiveTree->supportedKeySystem_.c_str());
+    SetSupportedDecrypterURN(m_adaptiveTree->m_supportedKeySystem);
+    LOG::Log(LOGDEBUG, "Supported URN: %s", m_adaptiveTree->m_supportedKeySystem.c_str());
   }
 
   // Preinitialize the DRM, if pre-initialisation data are provided
-  std::map<std::string, std::string> additionalHeaders = std::map<std::string, std::string>();
+  std::map<std::string, std::string> addHeaders;
   bool isSessionOpened{false};
 
   if (!m_kodiProps.m_drmPreInitData.empty())
@@ -233,8 +247,8 @@ bool CSession::Initialize()
     // used to make licensed manifest requests (via proxy callback)
     if (PreInitializeDRM(challengeB64, sessionId, isSessionOpened))
     {
-      additionalHeaders["challengeB64"] = STRING::URLEncode(challengeB64);
-      additionalHeaders["sessionId"] = sessionId;
+      addHeaders["challengeB64"] = STRING::URLEncode(challengeB64);
+      addHeaders["sessionId"] = sessionId;
     }
     else
     {
@@ -245,8 +259,10 @@ bool CSession::Initialize()
   // Open manifest file with location redirect support  bool mpdSuccess;
   std::string manifestUrl =
       m_adaptiveTree->location_.empty() ? m_manifestUrl : m_adaptiveTree->location_;
-  if (!m_adaptiveTree->open(manifestUrl, m_kodiProps.m_manifestUpdateParam, additionalHeaders) ||
-      m_adaptiveTree->empty())
+
+  m_adaptiveTree->SetManifestUpdateParam(manifestUrl, m_kodiProps.m_manifestUpdateParam);
+
+  if (!m_adaptiveTree->open(manifestUrl, addHeaders) || m_adaptiveTree->empty())
   {
     LOG::Log(LOGERROR, "Could not open / parse manifest (%s)", manifestUrl.c_str());
     return false;
@@ -257,9 +273,7 @@ bool CSession::Initialize()
       m_adaptiveTree->periods_.size(), m_adaptiveTree->current_period_->adaptationSets_.size(),
       m_adaptiveTree->has_timeshift_buffer_ ? "live" : "VOD");
 
-  // Always need at least 16s delay from live
-  if (m_adaptiveTree->live_delay_ < 16)
-    m_adaptiveTree->live_delay_ = 16;
+  m_adaptiveTree->PostOpen(m_kodiProps);
 
   return InitializePeriod(isSessionOpened);
 }
@@ -428,46 +442,26 @@ bool CSession::InitializeDRM(bool addDefaultKID /* = false */)
         return false;
       }
     }
-    std::string strkey(m_adaptiveTree->supportedKeySystem_.substr(9));
+    std::string strkey(m_adaptiveTree->m_supportedKeySystem.substr(9));
     size_t pos;
     while ((pos = strkey.find('-')) != std::string::npos)
       strkey.erase(pos, 1);
     if (strkey.size() != 32)
     {
-      LOG::Log(LOGERROR, "Key system mismatch (%s)!", m_adaptiveTree->supportedKeySystem_.c_str());
+      LOG::Log(LOGERROR, "Key system mismatch (%s)!", m_adaptiveTree->m_supportedKeySystem.c_str());
       return false;
     }
 
     unsigned char key_system[16];
     AP4_ParseHex(strkey.c_str(), key_system, 16);
-    uint32_t currentSessionTypes = 0;
 
     // cdmSession 0 is reserved for unencrypted streams
     for (size_t ses{1}; ses < m_cdmSessions.size(); ++ses)
     {
       AP4_DataBuffer init_data;
       const char* optionalKeyParameter{nullptr};
-      auto sessionPsshset{m_adaptiveTree->current_period_->psshSets_[ses]};
-      uint32_t sessionType = 0;
 
-      if (sessionPsshset.media_ > 0)
-      {
-        sessionType = sessionPsshset.media_;
-      }
-      else
-      {
-        switch (sessionPsshset.adaptation_set_->type_)
-        {
-          case adaptive::AdaptiveTree::VIDEO:
-            sessionType = adaptive::AdaptiveTree::Period::PSSH::MEDIA_VIDEO;
-            break;
-          case adaptive::AdaptiveTree::AUDIO:
-            sessionType = adaptive::AdaptiveTree::Period::PSSH::MEDIA_AUDIO;
-            break;
-          default:
-            break;
-        }
-      }
+      auto sessionPsshset{m_adaptiveTree->current_period_->psshSets_[ses]};
 
       if (sessionPsshset.pssh_ == "FILE")
       {
@@ -477,9 +471,8 @@ bool CSession::InitializeDRM(bool addDefaultKID /* = false */)
         {
           auto initialRepr{m_reprChooser->GetRepresentation(sessionPsshset.adaptation_set_)};
 
-          CStream stream{
-              *m_adaptiveTree, sessionPsshset.adaptation_set_,    initialRepr, m_mediaHeaders,
-              m_reprChooser,   m_kodiProps.m_playTimeshiftBuffer, false};
+          CStream stream{*m_adaptiveTree, sessionPsshset.adaptation_set_, initialRepr, m_kodiProps,
+                         false};
 
           stream.m_isEnabled = true;
           stream.m_adStream.start_stream();
@@ -548,22 +541,12 @@ bool CSession::InitializeDRM(bool addDefaultKID /* = false */)
         }
         else if (!sessionPsshset.defaultKID_.empty())
         {
-          init_data.SetData(reinterpret_cast<AP4_Byte*>(sessionPsshset.defaultKID_.data()), 16);
+          std::string licenseData = BASE64::Decode(m_kodiProps.m_licenseData);
+          // Replace KID placeholder, if any
+          STRING::ReplaceFirst(licenseData, "{KID}", sessionPsshset.defaultKID_);
 
-          std::string decLicenseData{BASE64::Decode(m_kodiProps.m_licenseData)};
-          uint8_t* decLicDataUInt{reinterpret_cast<uint8_t*>(decLicenseData.data())};
-
-          uint8_t* uuid{reinterpret_cast<uint8_t*>(strstr(decLicenseData.data(), "{KID}"))};
-          if (uuid)
-          {
-            memmove(uuid + 11, uuid, m_kodiProps.m_licenseData.size() - (uuid - decLicDataUInt));
-            memcpy(uuid, init_data.GetData(), init_data.GetDataSize());
-            init_data.SetData(decLicDataUInt, m_kodiProps.m_licenseData.size() + 11);
-          }
-          else
-          {
-            init_data.SetData(decLicDataUInt, m_kodiProps.m_licenseData.size());
-          }
+          init_data.SetData(reinterpret_cast<const AP4_Byte*>(licenseData.c_str()),
+                            static_cast<AP4_Size>(licenseData.size()));
         }
         else
           return false;
@@ -616,26 +599,33 @@ bool CSession::InitializeDRM(bool addDefaultKID /* = false */)
         std::string hexKid{StringUtils::ToHexadecimal(defaultKid)};
         LOG::Log(LOGDEBUG, "Initializing stream with KID: %s", hexKid.c_str());
 
-        // use shared ssd session if we already have 1 of the same stream type
-        if (currentSessionTypes & sessionType)
+        for (size_t i{1}; i < ses; ++i)
         {
-          for (size_t i{1}; i < ses; ++i)
+          if (m_decrypter->HasLicenseKey(m_cdmSessions[i].m_cencSingleSampleDecrypter, defkid))
           {
-            if (m_decrypter->HasLicenseKey(m_cdmSessions[i].m_cencSingleSampleDecrypter, defkid))
-            {
-              session.m_cencSingleSampleDecrypter = m_cdmSessions[i].m_cencSingleSampleDecrypter;
-              session.m_sharedCencSsd = true;
-              break;
-            }
+            session.m_cencSingleSampleDecrypter = m_cdmSessions[i].m_cencSingleSampleDecrypter;
+            session.m_sharedCencSsd = true;
+            break;
           }
         }
-      }
-      else if (defaultKid.empty() && !session.m_cencSingleSampleDecrypter)
-      {
-        LOG::Log(LOGWARNING, "Initializing stream with unknown KID!");
-      }
 
-      currentSessionTypes |= sessionType;
+      }
+      else if (defaultKid.empty())
+      {
+        for (size_t i{1}; i < ses; ++i)
+        {
+          if (sessionPsshset.pssh_ == m_adaptiveTree->current_period_->psshSets_[i].pssh_)
+          {
+            session.m_cencSingleSampleDecrypter = m_cdmSessions[i].m_cencSingleSampleDecrypter;
+            session.m_sharedCencSsd = true;
+            break;
+          }
+        }
+        if (!session.m_cencSingleSampleDecrypter)
+        {
+          LOG::Log(LOGWARNING, "Initializing stream with unknown KID!");
+        }
+      }
 
       if (m_decrypter && init_data.GetDataSize() >= 4 &&
           (session.m_cencSingleSampleDecrypter ||
@@ -776,8 +766,7 @@ void CSession::AddStream(adaptive::AdaptiveTree::AdaptationSet* adp,
                          bool isDefaultRepr,
                          uint32_t uniqueId)
 {
-  m_streams.push_back(std::make_unique<CStream>(*m_adaptiveTree, adp, initialRepr, m_mediaHeaders,
-                                                m_reprChooser, m_kodiProps.m_playTimeshiftBuffer,
+  m_streams.push_back(std::make_unique<CStream>(*m_adaptiveTree, adp, initialRepr, m_kodiProps,
                                                 m_firstPeriodInitialized));
 
   CStream& stream{*m_streams.back()};
@@ -1104,15 +1093,20 @@ void CSession::StartReader(
     CStream* stream, uint64_t seekTime, int64_t ptsDiff, bool preceeding, bool timing)
 {
   bool bReset = true;
+  ISampleReader* streamReader = stream->GetReader();
   if (timing)
+  {
     seekTime += stream->m_adStream.GetAbsolutePTSOffset();
+  }
   else
+  {
     seekTime -= ptsDiff;
+    streamReader->SetStartPTS(GetTimingStartPTS());
+  }
 
   stream->m_adStream.seek_time(static_cast<double>(seekTime / STREAM_TIME_BASE), preceeding,
                                bReset);
 
-  ISampleReader* streamReader = stream->GetReader();
   if (!streamReader)
   {
     LOG::LogF(LOGERROR, "Cannot get the stream reader");
@@ -1137,6 +1131,7 @@ bool CSession::GetNextSample(ISampleReader*& sampleReader)
 {
   CStream* res{nullptr};
   CStream* waiting{nullptr};
+  CStream* timingStream{GetTimingStream()};
 
   for (auto& stream : m_streams)
   {
@@ -1155,25 +1150,34 @@ bool CSession::GetNextSample(ISampleReader*& sampleReader)
         waiting = stream.get();
         break;
       }
-      else if (!streamReader->EOS() &&
-        AP4_SUCCEEDED(streamReader->Start(isStarted)))
+      else if (!streamReader->EOS())
       {
-        if (!res || streamReader->DTSorPTS() < res->GetReader()->DTSorPTS())
+        // Once the start PTS has been acquired for the timing stream, set this value
+        // to the other stream readers
+        if (stream.get() != timingStream &&
+            timingStream->GetReader()->GetStartPTS() != STREAM_NOPTS_VALUE &&
+            streamReader->GetStartPTS() == STREAM_NOPTS_VALUE)
         {
-          if (stream->m_adStream.waitingForSegment(true))
+          // want this to be the internal data's (not segment's) pts of 
+          // the first segment in period
+          streamReader->SetStartPTS(GetTimingStartPTS());
+        }
+        if (AP4_SUCCEEDED(streamReader->Start(isStarted)))
+        {
+          if (!res || streamReader->DTSorPTS() < res->GetReader()->DTSorPTS())
           {
-            waiting = stream.get();
-          }
-          else
-          {
-            res = stream.get();
+            if (stream->m_adStream.waitingForSegment(true))
+            {
+              waiting = stream.get();
+            }
+            else
+            {
+              res = stream.get();
+            }
           }
         }
       }
     }
-
-    if (isStarted && streamReader->GetInformation(stream->m_info))
-      m_changed = true;
   }
 
   if (waiting)
@@ -1184,8 +1188,6 @@ bool CSession::GetNextSample(ISampleReader*& sampleReader)
   {
     CheckFragmentDuration(*res);
     ISampleReader* sr{res->GetReader()};
-    if (sr->GetInformation(res->m_info))
-      m_changed = true;
 
     if (sr->PTS() != STREAM_NOPTS_VALUE)
       m_elapsedTime = PTSToElapsed(sr->PTS()) + GetChapterStartTime();
@@ -1243,7 +1245,7 @@ bool CSession::SeekTime(double seekTime, unsigned int streamId, bool preceeding)
       }
     }
 
-    maxSeek = (static_cast<double>(maxTime) / 1000) - m_adaptiveTree->live_delay_;
+    maxSeek = (static_cast<double>(maxTime) / 1000) - m_adaptiveTree->m_liveDelay;
     if (maxSeek < 0)
       maxSeek = 0;
 
@@ -1472,6 +1474,16 @@ int64_t CSession::GetChapterPos(int ch) const
   return sum / STREAM_TIME_BASE;
 }
 
+uint64_t CSession::GetTimingStartPTS() const
+{
+  if (CStream* timing = GetTimingStream())
+  {
+    return timing->GetReader()->GetStartPTS();
+  }
+  else
+    return 0;
+}
+
 uint64_t CSession::GetChapterStartTime() const
 {
   uint64_t start_time = 0;
@@ -1510,7 +1522,10 @@ bool CSession::SeekChapter(int ch)
   if (ch >= 0 && ch < static_cast<int>(m_adaptiveTree->periods_.size()) &&
       m_adaptiveTree->periods_[ch] != m_adaptiveTree->current_period_)
   {
-    m_adaptiveTree->next_period_ = m_adaptiveTree->periods_[ch];
+    auto newPd = m_adaptiveTree->periods_[ch];
+    m_adaptiveTree->next_period_ = newPd;
+    LOG::LogF(LOGDEBUG, "Switching to new Period (id=%s, start=%ld, seq=%d)", newPd->id_.c_str(),
+              newPd->start_, newPd->sequence_);
     for (auto& stream : m_streams)
     {
       ISampleReader* sr{stream->GetReader()};
