@@ -25,8 +25,11 @@
 #ifndef INPUTSTREAM_TEST_BUILD
 #include <kodi/Filesystem.h>
 #endif
+#include "kodi/tools/StringUtils.h"
 
 using namespace adaptive;
+using namespace std::chrono_literals;
+using namespace kodi::tools;
 using namespace UTILS;
 
 const size_t AdaptiveStream::MAXSEGMENTBUFFER = 10;
@@ -34,8 +37,7 @@ const size_t AdaptiveStream::MAXSEGMENTBUFFER = 10;
 AdaptiveStream::AdaptiveStream(AdaptiveTree& tree,
                                AdaptiveTree::AdaptationSet* adp,
                                AdaptiveTree::Representation* initialRepr,
-                               const std::map<std::string, std::string>& media_headers,
-                               bool play_timeshift_buffer,
+                               const UTILS::PROPERTIES::KodiProperties& kodiProps,
                                bool choose_rep)
   : thread_data_(nullptr),
     tree_(tree),
@@ -45,14 +47,15 @@ AdaptiveStream::AdaptiveStream(AdaptiveTree& tree,
     current_rep_(initialRepr),
     available_segment_buffers_(0),
     valid_segment_buffers_(0),
-    media_headers_(media_headers),
+    m_streamParams(kodiProps.m_streamParams),
+    m_streamHeaders(kodiProps.m_streamHeaders),
     segment_read_pos_(0),
     currentPTSOffset_(0),
     absolutePTSOffset_(0),
     lastUpdated_(std::chrono::system_clock::now()),
     m_fixateInitialization(false),
     m_segmentFileOffset(0),
-    play_timeshift_buffer_(play_timeshift_buffer),
+    play_timeshift_buffer_(kodiProps.m_playTimeshiftBuffer),
     choose_rep_(choose_rep),
     rep_counter_(1),
     prev_rep_(0),
@@ -66,7 +69,8 @@ AdaptiveStream::AdaptiveStream(AdaptiveTree& tree,
 
 AdaptiveStream::~AdaptiveStream()
 {
-  stop();
+  Stop();
+  DisposeWorker();
   clear();
 }
 
@@ -99,8 +103,7 @@ void AdaptiveStream::ResetActiveBuffer(bool oneValid)
   segment_read_pos_ = 0;
 }
 
-// Make sure worker is in CV wait state.
-void AdaptiveStream::StopWorker(STATE state)
+bool AdaptiveStream::StopWorker(STATE state)
 {
   // stop downloading chunks
   state_ = state;
@@ -108,20 +111,38 @@ void AdaptiveStream::StopWorker(STATE state)
   // make sure download section in worker thread is done.
   std::unique_lock<std::mutex> lckrw(thread_data_->mutex_rw_);
   while (worker_processing_)
+  {
+    // While we are waiting the state of worker may be changed
     thread_data_->signal_rw_.wait(lckrw);
-  // Now the worker thread should keep the lock until it starts waiting
-  // to get CV signaled - make sure we are at this point.
+  }
+
+  // Now if the state set is PAUSED/STOPPED the worker thread should keep the lock to mutex_dl_
+  // and wait for a signal to condition varibale "signal_dl_.wait",
+  // if state will be not changed to RUNNING next downloads will be not performed.
+
+  // Check if the worker state is changed by other situations
+  // e.g. stop playback or download cancelled
+  // that invalidated our status
+  return state_ == state;
+}
+
+void adaptive::AdaptiveStream::WaitWorker()
+{
+  // If the worker is in PAUSE/STOP state
+  // we wait here until condition variable "signal_dl_.wait" is executed,
+  // after that the worker will be waiting for a signal to unlock "signal_dl_.wait" (blocking thread)
   std::lock_guard<std::mutex> lckdl(thread_data_->mutex_dl_);
-  // Make sure that worker continues at next notify
+  // Make sure that worker continue the loop (avoid signal_dl_.wait block again the thread)
+  // and allow new downloads
   state_ = RUNNING;
 }
 
-bool AdaptiveStream::download_segment()
+bool AdaptiveStream::download_segment(const DownloadInfo& downloadInfo)
 {
-  if (download_url_.empty())
+  if (downloadInfo.m_url.empty())
     return false;
 
-  return download(download_url_, download_headers_, nullptr);
+  return download(downloadInfo, nullptr);
 }
 
 void AdaptiveStream::worker()
@@ -133,48 +154,73 @@ void AdaptiveStream::worker()
   {
     while (!thread_data_->thread_stop_ &&
            (state_ != RUNNING || valid_segment_buffers_ >= available_segment_buffers_))
+    {
       thread_data_->signal_dl_.wait(lckdl);
+    }
 
     if (!thread_data_->thread_stop_)
     {
       worker_processing_ = true;
 
-      prepareNextDownload();
+      DownloadInfo downloadInfo;
+      if (!prepareNextDownload(downloadInfo))
+      {
+        worker_processing_ = false;
+        continue;
+      }
 
       // tell the main thread that we have processed prepare_download;
       thread_data_->signal_dl_.notify_one();
       lckdl.unlock();
 
-      bool ret(download_segment());
-      unsigned int retryCount(10);
+      bool isLive = tree_.has_timeshift_buffer_;
 
-      //Some streaming software offers subtitle tracks with missing fragments, usually live tv
-      //When a programme is broadcasted that has subtitles, subtitles fragments are offered
-      //TODO: Ensure we continue with the next segment after one retry on errors
-      if (current_adp_->type_ == AdaptiveTree::SUBTITLE)
-        retryCount = 1;
+      //! @todo: for live content we should calculate max attempts and sleep timing
+      //! based on segment duration / playlist updates timing
+      size_t maxAttempts = isLive ? 10 : 6;
+      std::chrono::milliseconds msSleep = isLive ? 1000ms : 500ms;
 
-      while (!ret && state_ == RUNNING && retryCount-- && tree_.has_timeshift_buffer_)
+      //! @todo: Some streaming software offers subtitle tracks with missing fragments, usually live tv
+      //! When a programme is broadcasted that has subtitles, subtitles fragments are offered,
+      //! Ensure we continue with the next segment after one retry on errors
+      if (current_adp_->type_ == AdaptiveTree::SUBTITLE && isLive)
+        maxAttempts = 2;
+
+      size_t downloadAttempts = 1;
+      bool isSegmentDownloaded = false;
+
+      // Download errors may occur e.g. due to unstable connection, server overloading, ...
+      // then we try downloading the segment more times before aborting playback
+      while (state_ != STOPPED)
       {
-        std::this_thread::sleep_for(std::chrono::seconds(1));
-        LOG::LogF(LOGDEBUG, "Trying to reload segment ...");
-        ret = download_segment();
+        isSegmentDownloaded = download_segment(downloadInfo);
+        if (isSegmentDownloaded || downloadAttempts == maxAttempts || state_ == STOPPED)
+          break;
+
+        //! @todo: forcing thread sleep block the thread also while the state_ / thread_stop_ change values
+        //! we have to interrupt the sleep when it happens
+        std::this_thread::sleep_for(msSleep);
+        downloadAttempts++;
+        LOG::Log(LOGWARNING, "Segment download failed, attempt %zu...", downloadAttempts);
       }
 
       lckdl.lock();
 
-      //Signal finished download
+      if (!isSegmentDownloaded)
       {
         std::lock_guard<std::mutex> lckrw(thread_data_->mutex_rw_);
-        download_url_.clear();
-        if (!ret)
-          state_ = STOPPED;
+        // Download cancelled or cannot download the file
+        state_ = STOPPED;
       }
-      worker_processing_ = false;
 
-      thread_data_->signal_rw_.notify_one();
+      // Signal finished download
+      worker_processing_ = false;
+      thread_data_->signal_rw_.notify_all();
     }
   } while (!thread_data_->thread_stop_);
+
+  worker_processing_ = false;
+  lckdl.unlock();
 }
 
 int AdaptiveStream::SecondsSinceUpdate() const
@@ -186,27 +232,38 @@ int AdaptiveStream::SecondsSinceUpdate() const
           .count());
 }
 
-bool AdaptiveStream::download(const std::string& url,
-                              const std::map<std::string, std::string>& mediaHeaders,
+bool AdaptiveStream::download(const DownloadInfo& downloadInfo,
                               std::string* lockfreeBuffer)
 {
-  kodi::vfs::CFile file;
+  std::string url{downloadInfo.m_url};
 
-  // open the file
+  // Merge additional headers to the predefined one
+  std::map<std::string, std::string> headers = m_streamHeaders;
+  headers.insert(downloadInfo.m_addHeaders.begin(), downloadInfo.m_addHeaders.end());
+
+  // Append stream parameters, only if not already provided
+  if (url.find('?') == std::string::npos)
+    URL::AppendParameters(url, m_streamParams);
+
+  // Open the file
+  kodi::vfs::CFile file;
   if (!file.CURLCreate(url))
     return false;
+
   file.CURLAddOption(ADDON_CURL_OPTION_PROTOCOL, "seekable", "0");
   file.CURLAddOption(ADDON_CURL_OPTION_PROTOCOL, "acceptencoding", "gzip, deflate");
-  if (mediaHeaders.find("connection") == mediaHeaders.end())
+  if (headers.find("connection") == headers.end())
     file.CURLAddOption(ADDON_CURL_OPTION_HEADER, "connection", "keep-alive");
+
   file.CURLAddOption(ADDON_CURL_OPTION_PROTOCOL, "failonerror", "false");
 
-  for (const auto& entry : mediaHeaders)
+  for (auto& header : headers)
   {
-    file.CURLAddOption(ADDON_CURL_OPTION_HEADER, entry.first.c_str(), entry.second.c_str());
+    file.CURLAddOption(ADDON_CURL_OPTION_HEADER, header.first.c_str(), header.second.c_str());
   }
 
-  if (!file.CURLOpen(ADDON_READ_CHUNKED | ADDON_READ_NO_CACHE | ADDON_READ_AUDIO_VIDEO)) {
+  if (!file.CURLOpen(ADDON_READ_CHUNKED | ADDON_READ_NO_CACHE | ADDON_READ_AUDIO_VIDEO))
+  {
     LOG::Log(LOGERROR, "CURLOpen returned an error, download failed: %s", url.c_str());
     return false;
   }
@@ -253,10 +310,11 @@ bool AdaptiveStream::download(const std::string& url,
       else
       {
         // Store the data
-        // We only set lastChunk to true in the case of non-chunked transfers, the 
+        // We only set lastChunk to true in the case of non-chunked transfers, the
         // current structure does not allow for knowing the file has finished for
         // chunked transfers here - AtEnd() will return true while doing chunked transfers
-        if (write_data(bufferData.data(), byteRead, lockfreeBuffer, (!isChunked && file.AtEnd())))
+        if (write_data(bufferData.data(), byteRead, lockfreeBuffer, (!isChunked && file.AtEnd()),
+                       downloadInfo))
         {
           totalReadBytes += byteRead;
         }
@@ -426,7 +484,8 @@ bool AdaptiveStream::parseIndexRange(AdaptiveTree::Representation* rep, const st
 bool AdaptiveStream::write_data(const void* buffer,
                                 size_t buffer_size,
                                 std::string* lockfreeBuffer,
-                                bool lastChunk)
+                                bool lastChunk,
+                                const DownloadInfo& downloadInfo)
 {
   if (lockfreeBuffer)
   {
@@ -447,11 +506,11 @@ bool AdaptiveStream::write_data(const void* buffer,
 
     size_t insertPos(segment_buffer.size());
     segment_buffer.resize(insertPos + buffer_size);
-    tree_.OnDataArrived(download_segNum_, download_pssh_set_, m_iv,
-                        reinterpret_cast<const uint8_t*>(buffer),
-                        segment_buffer, insertPos, buffer_size, lastChunk);
+    tree_.OnDataArrived(downloadInfo.m_segmentNumber, downloadInfo.m_psshSet, m_decrypterIv,
+                        reinterpret_cast<const uint8_t*>(buffer), segment_buffer, insertPos,
+                        buffer_size, lastChunk);
   }
-  thread_data_->signal_rw_.notify_one();
+  thread_data_->signal_rw_.notify_all();
   return true;
 }
 
@@ -533,9 +592,11 @@ bool AdaptiveStream::start_stream()
         uint64_t duration(current_rep_->get_segment(pos)->startPTS_ -
                           current_rep_->get_segment(pos - 1)->startPTS_);
         size_t segmentPos{0};
-        if (pos > (tree_.live_delay_ * current_rep_->timescale_) / duration)
+        size_t segPosDelay =
+            static_cast<size_t>((tree_.m_liveDelay * current_rep_->timescale_) / duration);
+        if (pos > segPosDelay)
         {
-          segmentPos = pos - ((tree_.live_delay_ * current_rep_->timescale_) / duration);
+          segmentPos = pos - segPosDelay;
         }
         current_rep_->current_segment_ = current_rep_->get_segment(segmentPos);
       }
@@ -572,6 +633,7 @@ bool AdaptiveStream::start_stream()
   if (loadingSeg)
   {
     StopWorker(PAUSED);
+    WaitWorker();
 
     if (available_segment_buffers_)
       std::rotate(segment_buffers_.rend() - (available_segment_buffers_ + 1),
@@ -590,7 +652,8 @@ bool AdaptiveStream::start_stream()
     size_t valid_segment_buffers = valid_segment_buffers_;
     valid_segment_buffers_ = 0;
 
-    if (!prepareNextDownload() || !download_segment())
+    DownloadInfo downloadInfo;
+    if (!prepareNextDownload(downloadInfo) || !download_segment(downloadInfo))
       state_ = STOPPED;
 
     valid_segment_buffers_ = valid_segment_buffers + 1;
@@ -640,7 +703,7 @@ void AdaptiveStream::ReplacePlaceholder(std::string& url, const std::string plac
   url.replace(np - lenReplace, npe - np + lenReplace + 1, rangebuf);
 }
 
-bool AdaptiveStream::prepareNextDownload()
+bool AdaptiveStream::prepareNextDownload(DownloadInfo& downloadInfo)
 {
   // We assume, that we find the next segment to load in the next valid_segment_buffers_
   if (valid_segment_buffers_ >= available_segment_buffers_)
@@ -648,22 +711,25 @@ bool AdaptiveStream::prepareNextDownload()
 
   const AdaptiveTree::Representation* rep = segment_buffers_[valid_segment_buffers_].rep;
   const AdaptiveTree::Segment* seg = &segment_buffers_[valid_segment_buffers_].segment;
+
   // segNum == ~0U is initialization segment!
   uint64_t segNum = segment_buffers_[valid_segment_buffers_].segment_number;
   segment_buffers_[valid_segment_buffers_].buffer.clear();
   ++valid_segment_buffers_;
 
-  return prepareDownload(rep, seg, segNum);
+  return prepareDownload(rep, seg, segNum, downloadInfo);
 }
 
 bool AdaptiveStream::prepareDownload(const AdaptiveTree::Representation* rep,
                                      const AdaptiveTree::Segment* seg,
-                                     uint64_t segNum)
+                                     uint64_t segNum,
+                                     DownloadInfo& downloadInfo)
 {
   if (!seg)
     return false;
 
-  char rangebuf[128], *rangeHeader(0);
+  std::string rangeHeader;
+  std::string streamUrl;
 
   if (!(rep->flags_ & AdaptiveTree::Representation::SEGMENTBASE))
   {
@@ -672,64 +738,67 @@ bool AdaptiveStream::prepareDownload(const AdaptiveTree::Representation* rep,
       if (rep->flags_ & AdaptiveTree::Representation::URLSEGMENTS)
       {
         if (URL::IsUrlAbsolute(seg->url))
-          download_url_ = seg->url;
+          streamUrl = seg->url;
         else
-          download_url_ = URL::Join(rep->url_, seg->url);
+          streamUrl = URL::Join(rep->url_, seg->url);
       }
       else
-        download_url_ = rep->url_;
+        streamUrl = rep->url_;
       if (~seg->range_begin_)
       {
         uint64_t fileOffset = ~segNum ? m_segmentFileOffset : 0;
         if (~seg->range_end_)
-          sprintf(rangebuf, "bytes=%" PRIu64 "-%" PRIu64, seg->range_begin_ + fileOffset,
-                  seg->range_end_ + fileOffset);
+        {
+          rangeHeader = StringUtils::Format("bytes=%llu-%llu", seg->range_begin_ + fileOffset,
+                                            seg->range_end_ + fileOffset);
+        }
         else
-          sprintf(rangebuf, "bytes=%" PRIu64 "-", seg->range_begin_ + fileOffset);
-        rangeHeader = rangebuf;
+        {
+          rangeHeader = StringUtils::Format("bytes=%llu-", seg->range_begin_ + fileOffset);
+        }
       }
     }
     else if (~segNum) //templated segment
     {
-      download_url_ = rep->segtpl_.media_url;
-      ReplacePlaceholder(download_url_, "$Number", seg->range_end_);
-      ReplacePlaceholder(download_url_, "$Time", seg->range_begin_);
+      streamUrl = rep->segtpl_.media_url;
+      ReplacePlaceholder(streamUrl, "$Number", seg->range_end_);
+      ReplacePlaceholder(streamUrl, "$Time", seg->range_begin_);
     }
     else //templated initialization segment
-      download_url_ = rep->url_;
+      streamUrl = rep->url_;
   }
   else
   {
     if (rep->flags_ & AdaptiveTree::Representation::TEMPLATE && ~segNum)
     {
-      download_url_ = rep->segtpl_.media_url;
-      ReplacePlaceholder(download_url_, "$Number", rep->startNumber_);
-      ReplacePlaceholder(download_url_, "$Time", 0);
+      streamUrl = rep->segtpl_.media_url;
+      ReplacePlaceholder(streamUrl, "$Number", rep->startNumber_);
+      ReplacePlaceholder(streamUrl, "$Time", 0);
     }
     else
-      download_url_ = rep->url_;
+      streamUrl = rep->url_;
+
     if (~seg->range_begin_)
     {
       uint64_t fileOffset = ~segNum ? m_segmentFileOffset : 0;
       if (~seg->range_end_)
-        sprintf(rangebuf, "bytes=%" PRIu64 "-%" PRIu64, seg->range_begin_ + fileOffset,
-                seg->range_end_ + fileOffset);
+      {
+        rangeHeader = StringUtils::Format("bytes=%llu-%llu", seg->range_begin_ + fileOffset,
+                                          seg->range_end_ + fileOffset);
+      }
       else
-        sprintf(rangebuf, "bytes=%" PRIu64 "-", seg->range_begin_ + fileOffset);
-      rangeHeader = rangebuf;
+      {
+        rangeHeader = StringUtils::Format("bytes=%llu-", seg->range_begin_ + fileOffset);
+      }
     }
   }
 
-  download_segNum_ = segNum;
-  download_pssh_set_ = seg->pssh_set_;
-  download_headers_ = media_headers_;
-  if (rangeHeader)
-    download_headers_["Range"] = rangeHeader;
-  else
-    download_headers_.erase("Range");
-
-  download_url_ = tree_.BuildDownloadUrl(download_url_);
+  downloadInfo.m_segmentNumber = segNum;
+  downloadInfo.m_psshSet = seg->pssh_set_;
+  if (!rangeHeader.empty())
+    downloadInfo.m_addHeaders["Range"] = rangeHeader; 
 
+  downloadInfo.m_url = tree_.BuildDownloadUrl(streamUrl);
   return true;
 }
 
@@ -746,9 +815,9 @@ bool AdaptiveStream::ensureSegment()
     // wait until worker is ready for new segment
     std::unique_lock<std::mutex> lck(thread_data_->mutex_dl_);
     // lock live segment updates
-    std::lock_guard<std::mutex> lckTree(tree_.GetTreeMutex());
+    std::lock_guard<adaptive::AdaptiveTree::TreeUpdateThread> lckUpdTree(tree_.GetTreeUpdMutex());
 
-    if (tree_.HasUpdateThread() && SecondsSinceUpdate() > 1)
+    if (tree_.HasManifestUpdates() && SecondsSinceUpdate() > 1)
     {
       tree_.RefreshSegments(current_period_, current_adp_, current_rep_, current_adp_->type_);
       lastUpdated_ = std::chrono::system_clock::now();
@@ -881,10 +950,13 @@ bool AdaptiveStream::ensureSegment()
         return false;
       }
     }
-    else if (tree_.HasUpdateThread() && current_period_ == tree_.periods_.back())
+    else if (tree_.HasManifestUpdates() && current_period_ == tree_.periods_.back())
     {
-      current_rep_->flags_ |= AdaptiveTree::Representation::WAITFORSEGMENT;
-      LOG::LogF(LOGDEBUG, "Begin WaitForSegment stream %s", current_rep_->id.c_str());
+      if ((current_rep_->flags_ & AdaptiveTree::Representation::WAITFORSEGMENT) == 0)
+      {
+        current_rep_->flags_ |= AdaptiveTree::Representation::WAITFORSEGMENT;
+        LOG::LogF(LOGDEBUG, "Begin WaitForSegment stream %s", current_rep_->id.c_str());
+      }
       std::this_thread::sleep_for(std::chrono::milliseconds(100));
       return false;
     }
@@ -968,20 +1040,12 @@ bool AdaptiveStream::retrieveCurrentSegmentBufferSize(size_t& size)
   if (state_ == STOPPED)
     return false;
 
-  std::unique_lock<std::mutex> lckrw(thread_data_->mutex_rw_);
-
-  while (true)
-  {
-    if (worker_processing_)
-    {
-      thread_data_->signal_rw_.wait(lckrw);
-      continue;
-    }
-    size = segment_buffers_[0].buffer.size();
-    return true;
-  }
-
-  return false;
+  if (!StopWorker(PAUSED))
+    return false;
+  
+  size = segment_buffers_[0].buffer.size();
+  WaitWorker();
+  return true;
 }
 
 uint64_t AdaptiveStream::getMaxTimeMs()
@@ -1007,6 +1071,17 @@ uint64_t AdaptiveStream::getMaxTimeMs()
   return (timeExt - absolutePTSOffset_) / 1000;
 }
 
+void AdaptiveStream::ResetCurrentSegment(const AdaptiveTree::Segment* newSegment)
+{
+  StopWorker(STOPPED);
+  WaitWorker();
+  // EnsureSegment loads always the next segment, so go back 1
+  current_rep_->current_segment_ =
+    current_rep_->get_segment(current_rep_->get_segment_pos(newSegment) - 1);
+  // TODO: if new segment is already prefetched, don't ResetActiveBuffer;
+  ResetActiveBuffer(false);
+}
+
 bool AdaptiveStream::seek_time(double seek_seconds, bool preceeding, bool& needReset)
 {
   if (!current_rep_)
@@ -1015,7 +1090,7 @@ bool AdaptiveStream::seek_time(double seek_seconds, bool preceeding, bool& needR
   if (current_rep_->flags_ & AdaptiveTree::Representation::SUBTITLESTREAM)
     return true;
 
-  std::unique_lock<std::mutex> lckTree(tree_.GetTreeMutex());
+  std::lock_guard<adaptive::AdaptiveTree::TreeUpdateThread> lckUpdTree(tree_.GetTreeUpdMutex());
 
   uint32_t choosen_seg(~0);
 
@@ -1057,29 +1132,24 @@ bool AdaptiveStream::seek_time(double seek_seconds, bool preceeding, bool& needR
   if (newSeg)
   {
     needReset = true;
-    if ((old_seg && newSeg != old_seg) || (!preceeding && state_ == STOPPED))
+    if (newSeg != old_seg)
     {
-      StopWorker(STOPPED);
-      // EnsureSegment loads always the next segment, so go back 1
-      current_rep_->current_segment_ =
-          current_rep_->get_segment(current_rep_->get_segment_pos(newSeg) - 1);
-      // TODO: if new segment is already prefetched, don't ResetActiveBuffer;
-      ResetActiveBuffer(false);
-      if (newSeg == old_seg && !preceeding)
-      {
-        absolute_position_ -= segment_read_pos_;
-        segment_read_pos_ = 0;
-      }
+      ResetCurrentSegment(newSeg);
     }
-    else if (!preceeding && !old_seg)
+    else if (!preceeding)
     {
+      // restart stream if it has 'finished', e.g in the case of subtitles
+      // where there may be a few or only one segment for the period and 
+      // the stream is now in EOS state (all data already passed to Kodi)
+      if (state_ == STOPPED)
+      {
+        ResetCurrentSegment(newSeg);
+      }
       absolute_position_ -= segment_read_pos_;
       segment_read_pos_ = 0;
     }
-    else if (preceeding)
-    {
+    else
       needReset = false;
-    }
     return true;
   }
   else
@@ -1089,9 +1159,10 @@ bool AdaptiveStream::seek_time(double seek_seconds, bool preceeding, bool& needR
 
 bool AdaptiveStream::waitingForSegment(bool checkTime) const
 {
-  if (tree_.HasUpdateThread())
+  if (tree_.HasManifestUpdates() && state_ == RUNNING)
   {
-    std::lock_guard<std::mutex> lckTree(tree_.GetTreeMutex());
+    std::lock_guard<adaptive::AdaptiveTree::TreeUpdateThread> lckUpdTree(tree_.GetTreeUpdMutex());
+
     if (current_rep_ && (current_rep_->flags_ & AdaptiveTree::Representation::WAITFORSEGMENT) != 0)
       return !checkTime ||
              (current_adp_->type_ != AdaptiveTree::VIDEO &&
@@ -1128,8 +1199,10 @@ bool AdaptiveStream::ResolveSegmentBase(AdaptiveTree::Representation* rep, bool
       return false;
 
     std::string sidxBuffer;
-    if (prepareDownload(rep, &seg, segNum) &&
-        download(download_url_, download_headers_, &sidxBuffer) && parseIndexRange(rep, sidxBuffer))
+    DownloadInfo downloadInfo;
+
+    if (prepareDownload(rep, &seg, segNum, downloadInfo) && download(downloadInfo, &sidxBuffer) &&
+        parseIndexRange(rep, sidxBuffer))
     {
       const_cast<AdaptiveTree::Representation*>(rep)->flags_ &=
           ~AdaptiveTree::Representation::SEGMENTBASE;
@@ -1148,7 +1221,7 @@ void AdaptiveStream::info(std::ostream& s)
     << " bandwidth: " << current_rep_->bandwidth_ << std::endl;
 }
 
-void AdaptiveStream::stop()
+void AdaptiveStream::Stop()
 {
   if (current_rep_) 
   {
@@ -1158,14 +1231,33 @@ void AdaptiveStream::stop()
 
   if (thread_data_)
   {
+    thread_data_->Stop();
     StopWorker(STOPPED);
-    delete thread_data_;
-    thread_data_ = nullptr;
   }
-};
+}
 
 void AdaptiveStream::clear()
 {
   current_adp_ = 0;
   current_rep_ = 0;
 }
+
+void adaptive::AdaptiveStream::DisposeWorker()
+{
+  if (thread_data_)
+  {
+    if (worker_processing_)
+    {
+      LOG::LogF(LOGERROR,
+                "Cannot delete worker thread, download is in progress.");
+      return;
+    }
+    if (!thread_data_->thread_stop_)
+    {
+      LOG::LogF(LOGERROR, "Cannot delete worker thread, loop is still running.");
+      return;
+    }
+    delete thread_data_;
+    thread_data_ = nullptr;
+  }
+}
