@@ -9,7 +9,9 @@
 #include "AdaptiveTree.h"
 #include "Chooser.h"
 
+#include "../utils/FileUtils.h"
 #include "../utils/UrlUtils.h"
+#include "../utils/Utils.h"
 #include "../utils/log.h"
 
 #ifndef INPUTSTREAM_TEST_BUILD
@@ -18,6 +20,7 @@
 #endif
 
 #include <algorithm>
+// #include <cassert>
 #include <chrono>
 #include <stdlib.h>
 #include <string.h>
@@ -43,29 +46,32 @@ namespace adaptive
     *this = *src;
   }
 
-  AdaptiveTree::AdaptiveTree(const UTILS::PROPERTIES::KodiProperties& kodiProps,
-                             CHOOSER::IRepresentationChooser* reprChooser)
-    : m_kodiProps(kodiProps),
-      m_reprChooser(reprChooser),
-      m_streamHeaders(kodiProps.m_streamHeaders),
-      current_period_(nullptr),
-      next_period_(nullptr),
-      parser_(0),
-      currentNode_(0),
-      segcount_(0),
-      overallSeconds_(0),
-      stream_start_(0),
-      available_time_(0),
-      base_time_(0),
-      live_delay_(0),
-      minPresentationOffset(0),
-      has_timeshift_buffer_(false),
-      has_overall_seconds_(false),
-      updateInterval_(~0),
-      updateThread_(nullptr),
-      lastUpdated_(std::chrono::system_clock::now()),
-      m_cryptoMode(CryptoMode::NONE)
+  AdaptiveTree::AdaptiveTree(CHOOSER::IRepresentationChooser* reprChooser)
+  : m_reprChooser(reprChooser)
   {
+  }
+
+  AdaptiveTree::AdaptiveTree(const AdaptiveTree& left) : AdaptiveTree(left.m_reprChooser)
+  {
+    m_manifestParams = left.m_manifestParams;
+    m_manifestHeaders = left.m_manifestHeaders;
+    m_settings = left.m_settings;
+    m_supportedKeySystem = left.m_supportedKeySystem;
+    m_pathSaveManifest = left.m_pathSaveManifest;
+  }
+
+  void AdaptiveTree::Configure(const UTILS::PROPERTIES::KodiProperties& kodiProps)
+  {
+    if (kodi::addon::GetSettingBoolean("debug.save.manifest"))
+    {
+      m_pathSaveManifest = FILESYS::PathCombine(FILESYS::GetAddonUserPath(), "manifests");
+      // Delete previously saved manifest files
+      FILESYS::RemoveDirectory(m_pathSaveManifest, false);
+    }
+
+    m_manifestParams = kodiProps.m_manifestParams;
+    m_manifestHeaders = kodiProps.m_manifestHeaders;
+
     // Convenience way to share common addon settings we avoid
     // calling the API many times to improve parsing performance
     m_settings.m_bufferAssuredDuration =
@@ -74,22 +80,17 @@ namespace adaptive
         static_cast<uint32_t>(kodi::addon::GetSettingInt("MAXBUFFERDURATION"));
   }
 
-  AdaptiveTree::~AdaptiveTree()
+  void AdaptiveTree::PostOpen(const UTILS::PROPERTIES::KodiProperties& kodiProps)
   {
-    has_timeshift_buffer_ = false;
-    if (updateThread_)
-    {
-      {
-        std::lock_guard<std::mutex> lck(updateMutex_);
-        updateVar_.notify_one();
-      }
-      updateThread_->join();
-      delete updateThread_;
-    }
+    // A manifest can provide live delay value, if not so we use our default
+    // value of 16 secs, this is needed to ensure an appropriate playback,
+    // an add-on can override the delay to try fix edge use cases
+    if (kodiProps.m_liveDelay >= 16)
+      m_liveDelay = kodiProps.m_liveDelay;
+    else if (m_liveDelay < 16)
+      m_liveDelay = 16;
 
-    std::lock_guard<std::mutex> lck(treeMutex_);
-    for (std::vector<Period*>::const_iterator bp(periods_.begin()), ep(periods_.end()); bp != ep; ++bp)
-      delete *bp;
+    StartUpdateThread();
   }
 
   void AdaptiveTree::FreeSegments(Period* period, Representation* rep)
@@ -127,7 +128,7 @@ namespace adaptive
 
   void AdaptiveTree::SetFragmentDuration(const AdaptationSet* adp, const Representation* rep, size_t pos, uint64_t timestamp, uint32_t fragmentDuration, uint32_t movie_timescale)
   {
-    if (!has_timeshift_buffer_ || HasUpdateThread() ||
+    if (!has_timeshift_buffer_ || HasManifestUpdates() ||
       (rep->flags_ & AdaptiveTree::Representation::URLSEGMENTS) != 0)
       return;
 
@@ -348,19 +349,6 @@ namespace adaptive
     return true;
   }
 
-  void AdaptiveTree::PrepareManifestUrl(const std::string &url, const std::string &manifestUpdateParam)
-  {
-    manifest_url_ = url;
-
-    if (manifestUpdateParam.empty())
-    {
-      update_parameter_ = URL::GetParametersFromPlaceholder(manifest_url_, "$START_NUMBER$");
-      manifest_url_.resize(manifest_url_.size() - update_parameter_.size());
-    }
-    else
-      update_parameter_ = manifestUpdateParam;
-  }
-
   std::string AdaptiveTree::BuildDownloadUrl(const std::string& url) const
   {
     if (URL::IsUrlAbsolute(url))
@@ -403,33 +391,29 @@ namespace adaptive
     }
   }
 
-  void AdaptiveTree::RefreshUpdateThread()
-  {
-    if (HasUpdateThread())
-    {
-      std::lock_guard<std::mutex> lck(updateMutex_);
-      updateVar_.notify_one();
-    }
-  }
-
   void AdaptiveTree::StartUpdateThread()
   {
-    if (!updateThread_ && ~updateInterval_ && has_timeshift_buffer_ && !update_parameter_.empty())
-      updateThread_ = new std::thread(&AdaptiveTree::SegmentUpdateWorker, this);
+    if (HasManifestUpdates())
+      m_updThread.Initialize(this);
   }
 
-  void AdaptiveTree::SegmentUpdateWorker()
+  bool AdaptiveTree::DownloadManifest(std::string url,
+                                      const std::map<std::string, std::string>& addHeaders,
+                                      std::stringstream& data,
+                                      HTTPRespHeaders& respHeaders)
   {
-    std::unique_lock<std::mutex> updLck(updateMutex_);
-    while (~updateInterval_ && has_timeshift_buffer_)
+    std::map<std::string, std::string> manifestHeaders = m_manifestHeaders;
+    // Merge additional headers to the predefined one
+    for (auto& headerIt : addHeaders)
     {
-      if (updateVar_.wait_for(updLck, std::chrono::milliseconds(updateInterval_)) == std::cv_status::timeout)
-      {
-        std::lock_guard<std::mutex> lck(treeMutex_);
-        lastUpdated_ = std::chrono::system_clock::now();
-        RefreshLiveSegments();
-      }
+      manifestHeaders[headerIt.first] = headerIt.second;
     }
+
+    // Append manifest parameters, only if not already provided (e.g. manifest update)
+    if (url.find('?') == std::string::npos)
+      URL::AppendParameters(url, m_manifestParams);
+
+    return download(url, manifestHeaders, data, respHeaders);
   }
 
   bool AdaptiveTree::download(const std::string& url,
@@ -526,4 +510,89 @@ namespace adaptive
     return false;
   }
 
-} // namespace
+  void AdaptiveTree::SaveManifest(const std::string& fileNameSuffix,
+                                  const std::stringstream& data,
+                                  std::string_view info)
+  {
+    if (!m_pathSaveManifest.empty())
+    {
+      // We create a filename based on current timestamp
+      // to allow files to be kept in download order useful for live streams
+      std::string filename = "manifest_" + std::to_string(UTILS::GetTimestamp());
+      if (!fileNameSuffix.empty())
+        filename += "_" + fileNameSuffix;
+
+      filename += ".txt";
+      std::string filePath = FILESYS::PathCombine(m_pathSaveManifest, filename);
+
+      // Manage duplicate files and limit them, too many means a problem to be solved
+      if (FILESYS::CheckDuplicateFilePath(filePath, 10))
+      {
+        std::string dataToSave{info};
+        dataToSave += "\n\n";
+        dataToSave += data.str();
+
+        if (FILESYS::SaveFile(filePath, dataToSave, false))
+          LOG::Log(LOGDEBUG, "Manifest saved to: %s", filePath.c_str());
+      }
+    }
+  }
+
+  AdaptiveTree::TreeUpdateThread::~TreeUpdateThread()
+  {
+    // assert(m_waitQueue == 0); // Debug only, missing resume
+    m_threadStop = true;
+
+    if (m_thread.joinable())
+    {
+      m_cvUpdInterval.notify_all(); // Unlock possible waiting
+      m_thread.join();
+    }
+  }
+
+  void AdaptiveTree::TreeUpdateThread::Initialize(AdaptiveTree* tree)
+  {
+    if (!m_thread.joinable())
+    {
+      m_tree = tree;
+      m_thread = std::thread(&TreeUpdateThread::Worker, this);
+    }
+  }
+
+  void AdaptiveTree::TreeUpdateThread::Worker()
+  {
+    std::unique_lock<std::mutex> updLck(m_updMutex);
+
+    while (~m_tree->m_updateInterval && !m_threadStop)
+    {
+      if (m_cvUpdInterval.wait_for(updLck, std::chrono::milliseconds(m_tree->m_updateInterval)) ==
+          std::cv_status::timeout)
+      {
+        updLck.unlock();
+        // If paused, wait until last "Resume" will be called
+        std::unique_lock<std::mutex> lckWait(m_waitMutex);
+        m_cvWait.wait(lckWait, [&] { return m_waitQueue == 0; });
+
+        updLck.lock();
+        m_tree->RefreshLiveSegments();
+      }
+    }
+  }
+
+  void AdaptiveTree::TreeUpdateThread::Pause()
+  {
+    // If an update is already in progress the wait until its finished
+    std::lock_guard<std::mutex> updLck{m_updMutex};
+    m_waitQueue++;
+  }
+
+  void AdaptiveTree::TreeUpdateThread::Resume()
+  {
+    // assert(m_waitQueue != 0); // Debug only, resume without any pause
+    m_waitQueue--;
+    // If there are no more pauses, unblock the update thread
+    if (m_waitQueue == 0)
+      m_cvWait.notify_all();
+  }
+
+  } // namespace adaptive
