@@ -8,11 +8,12 @@
 
 #pragma once
 
-#include "../CryptoMode.h"
+#include "../utils/CryptoUtils.h"
 #include "../utils/PropertiesUtils.h"
 #include "expat.h"
 
 #include <algorithm>
+#include <atomic>
 #include <chrono>
 #include <condition_variable>
 #include <inttypes.h>
@@ -30,7 +31,7 @@
 #include <kodi/AddonBase.h>
 #endif
 
- // Forward namespace/class
+// Forward namespace/class
 namespace CHOOSER
 {
 class IRepresentationChooser;
@@ -478,25 +479,30 @@ public:
     bool need_secure_decoder_ = false;
     SPINCACHE<uint32_t> segment_durations_;
     SegmentTemplate segtpl_;
-  }*current_period_, *next_period_;
+  } *current_period_{nullptr}, *next_period_{nullptr};
 
   std::vector<Period*> periods_;
   std::string manifest_url_;
   std::string base_url_;
   std::string effective_url_;
-  std::string update_parameter_;
-  HTTPRespHeaders m_manifestHeaders;
+  std::string m_manifestUpdateParam;
 
   /* XML Parsing*/
-  XML_Parser parser_;
-  uint32_t currentNode_;
-  size_t segcount_;
+  XML_Parser parser_{nullptr};
+  uint32_t currentNode_{0};
+  size_t segcount_{0};
   uint32_t initial_sequence_ = ~0U;
-  uint64_t overallSeconds_, stream_start_, available_time_, base_time_, live_delay_;
-  uint64_t minPresentationOffset;
-  bool has_timeshift_buffer_, has_overall_seconds_;
+  uint64_t overallSeconds_{0};
+  uint64_t stream_start_{0};
+  uint64_t available_time_{0};
+  uint64_t base_time_{0};
+  uint64_t m_liveDelay{0};
+  uint64_t minPresentationOffset{0};
+  bool has_timeshift_buffer_{false};
+  bool has_overall_seconds_{false};
 
-  std::string supportedKeySystem_, location_;
+  std::string m_supportedKeySystem;
+  std::string location_;
 
   uint8_t adpChannelCount_, adp_pssh_set_;
   int adpwidth_;
@@ -510,17 +516,34 @@ public:
   std::string current_pssh_;
   std::string current_defaultKID_;
   std::string current_iv_;
-  CryptoMode m_cryptoMode;
+  CryptoMode m_cryptoMode{CryptoMode::NONE};
   std::string license_url_;
 
   std::string strXMLText_;
 
-  AdaptiveTree(const UTILS::PROPERTIES::KodiProperties& properties,
-    CHOOSER::IRepresentationChooser* reprChooser);
-  virtual ~AdaptiveTree();
+  AdaptiveTree(CHOOSER::IRepresentationChooser* reprChooser);
+  AdaptiveTree(const AdaptiveTree& left);
+  virtual ~AdaptiveTree()
+  {
+    for (auto period : periods_)
+      delete period;
+  }
+
+  /*!
+   * \brief Configure the adaptive tree.
+   * \param kodiProps The Kodi properties
+   */
+  virtual void Configure(const UTILS::PROPERTIES::KodiProperties& kodiProps);
+
+  virtual bool open(const std::string& url) = 0;
+  virtual bool open(const std::string& url, std::map<std::string, std::string> additionalHeaders) = 0;
+
+  /*!
+   * \brief Performs tasks after opening the manifest
+   * \param kodiProps The Kodi properties
+   */
+  virtual void PostOpen(const UTILS::PROPERTIES::KodiProperties& kodiProps);
 
-  virtual bool open(const std::string& url, const std::string& manifestUpdateParam) = 0;
-  virtual bool open(const std::string& url, const std::string& manifestUpdateParam, std::map<std::string, std::string> additionalHeaders) = 0;
   virtual PREPARE_RESULT prepareRepresentation(Period* period,
                                                AdaptationSet* adp,
                                                Representation* rep,
@@ -563,9 +586,12 @@ public:
 
   std::string BuildDownloadUrl(const std::string& url) const;
 
-  std::mutex &GetTreeMutex() { return treeMutex_; };
-  bool HasUpdateThread() const { return updateThread_ != 0 && has_timeshift_buffer_ && updateInterval_ && !update_parameter_.empty(); };
-  void RefreshUpdateThread();
+  bool HasManifestUpdates() const
+  {
+    return ~m_updateInterval && m_updateInterval > 0 && has_timeshift_buffer_ &&
+           !m_manifestUpdateParam.empty();
+  }
+
   const std::chrono::time_point<std::chrono::system_clock> GetLastUpdated() const { return lastUpdated_; };
 
   CHOOSER::IRepresentationChooser* GetRepChooser() { return m_reprChooser; }
@@ -579,9 +605,85 @@ public:
 
   virtual AdaptiveTree* Clone() const = 0;
 
+  /*!
+   * \brief Set the manifest update url parameter, used to force enabling manifest updates,
+   *        by default by set the @param argument to "full" value, but the behaviour
+   *        could change, it depends on the parser implementation.
+   * \param manifestUrl The original manifest url value may be modified,
+   *                    refer to each implementation of the parser.
+   * \param param The update parameter, by default is accepted "full" value,
+   *              refer to each implementation of the parser.
+   */
+  virtual void SetManifestUpdateParam(std::string& manifestUrl, std::string_view param)
+  {
+    m_manifestUpdateParam = param;
+  }
+
   Settings m_settings;
 
+  class TreeUpdateThread
+  {
+  public:
+    TreeUpdateThread() = default;
+    ~TreeUpdateThread();
+
+    void Initialize(AdaptiveTree* tree);
+
+    // \brief Reset start time (make exit the condition variable m_cvUpdInterval and re-start the timeout)
+    void ResetStartTime() { m_cvUpdInterval.notify_all(); }
+
+    // \brief As "std::mutex" lock, but put in pause the manifest updates (support std::lock_guard).
+    //        If an update is in progress, block the code until the update is finished.
+    void lock() { Pause(); }
+
+    // \brief As "std::mutex" unlock, but resume the manifest updates (support std::lock_guard).
+    void unlock() { Resume(); }
+
+  private:
+    void Worker();
+    void Pause();
+    void Resume();
+
+    std::thread m_thread;
+    AdaptiveTree* m_tree{nullptr};
+
+    // Reentrant mode to pause updates, which will be resumed as soon as the queue is removed,
+    // each time Pause() is called will add a wait queue
+    // each time Resume() is called remove a wait queue
+    // when there are no more wait queue, the updates will be resumed.
+    std::atomic<uint32_t> m_waitQueue{0};
+
+    std::mutex m_updMutex;
+    std::condition_variable m_cvUpdInterval;
+    std::mutex m_waitMutex;
+    std::condition_variable m_cvWait;
+    bool m_threadStop{false};
+  };
+
+  /*!
+   * \brief Get the tree manifest update thread to be used like an "std::mutex"
+   *        to put in pause/resume tree manifest updates betweeen other operations.
+   *        NOTE: this is a custom mutex that act as reentrant way,
+   *        then can be used at same time by different threads, the code that call the mutex lock
+   *        will be blocked only when the manifest updates are already in progress.
+   */
+  TreeUpdateThread& GetTreeUpdMutex() { return m_updThread; };
+
 protected:
+
+  /*!
+   * \brief Download manifest file.
+   * \param url The url of the file to download
+   * \param addHeaders Additional headers to add in the HTTP request
+   * \param data [OUT] Return the HTTP response data
+   * \param respHeaders [OUT] Return the HTTP response headers
+   * \return True if has success, otherwise false
+   */
+  virtual bool DownloadManifest(std::string url,
+                                const std::map<std::string, std::string>& addHeaders,
+                                std::stringstream& data,
+                                HTTPRespHeaders& respHeaders);
+
   /*!
    * \brief Download a file (At each call feed also the repr. chooser
             to calculate the initial bandwidth).
@@ -596,25 +698,33 @@ protected:
                         std::stringstream& data,
                         HTTPRespHeaders& respHeaders);
 
+  /*!
+   * \brief Save manifest data to a file for debugging purpose.
+   * \param fileNameSuffix Suffix to add to the filename generated.
+   * \param data The manifest data to save.
+   * \param info Additionals info to be add before the data.
+   */
+  virtual void SaveManifest(const std::string& fileNameSuffix,
+                            const std::stringstream& data,
+                            std::string_view info);
+
   bool PreparePaths(const std::string &url);
-  void PrepareManifestUrl(const std::string &url, const std::string &manifestUpdateParam);
   void SortTree();
 
   // Live segment update section
   virtual void StartUpdateThread();
-  virtual void RefreshLiveSegments(){};
+  virtual void RefreshLiveSegments() { lastUpdated_ = std::chrono::system_clock::now(); }
+  std::atomic<uint32_t> m_updateInterval{~0U};
+  TreeUpdateThread m_updThread;
+  std::atomic<std::chrono::time_point<std::chrono::system_clock>> lastUpdated_{
+      std::chrono::system_clock::now()};
 
-  uint32_t updateInterval_;
-  std::mutex treeMutex_, updateMutex_;
-  std::condition_variable updateVar_;
-  std::thread *updateThread_;
-  std::chrono::time_point<std::chrono::system_clock> lastUpdated_;
-  std::map<std::string, std::string> m_streamHeaders;
-  const UTILS::PROPERTIES::KodiProperties m_kodiProps;
+  std::string m_manifestParams;
+  std::map<std::string, std::string> m_manifestHeaders;
   CHOOSER::IRepresentationChooser* m_reprChooser{nullptr};
 
-private:
-  void SegmentUpdateWorker();
+  // Provide the path where the manifests will be saved, if debug enabled
+  std::string m_pathSaveManifest;
 };
 
 } // namespace adaptive
