@@ -65,20 +65,35 @@ bool CInputStreamAdaptive::Open(const kodi::addon::InputstreamProperty& props)
   if (m_kodiProps.m_isLicensePersistentStorage)
     drmConfig |= SSD::SSD_DECRYPTER::CONFIG_PERSISTENTSTORAGE;
 
-  std::map<std::string, std::string> mediaHeaders{m_kodiProps.m_streamHeaders};
-
-  // If the URL contains headers then replace the stream headers
-  std::string::size_type posHeader(url.find("|"));
+  // If the manifest URL contains headers then replace the manifest headers set with property
+  //! @todo: remove pipe support on Kodi v21
+  size_t posHeader = url.find("|");
   if (posHeader != std::string::npos)
   {
-    m_kodiProps.m_streamHeaders.clear();
-    ParseHeaderString(m_kodiProps.m_streamHeaders, url.substr(posHeader + 1));
+    LOG::Log(LOGWARNING, "Set headers to the manifest url by using pipe \"|\" char is deprecated, "
+                         "and will be removed in future.\n"
+                         "Use \"inputstream.adaptive.manifest_headers\" and "
+                         "\"inputstream.adaptive.stream_headers\" properties instead.");
+    m_kodiProps.m_manifestHeaders.clear();
+    ParseHeaderString(m_kodiProps.m_manifestHeaders, url.substr(posHeader + 1));
     url = url.substr(0, posHeader);
-  }
-  if (mediaHeaders.empty())
-    mediaHeaders = m_kodiProps.m_streamHeaders;
 
-  m_session = std::make_shared<CSession>(m_kodiProps, url, mediaHeaders, props.GetProfileFolder());
+    if (m_kodiProps.m_streamHeaders.empty())
+      m_kodiProps.m_streamHeaders = m_kodiProps.m_manifestHeaders;
+  }
+
+  //! @todo: remove this old forced behaviour on Kodi v21
+  if (m_kodiProps.m_manifestHeaders.empty() && !m_kodiProps.m_streamHeaders.empty())
+  {
+    LOG::Log(
+        LOGWARNING,
+        "Set headers to the manifest by using \"inputstream.adaptive.stream_headers\" property "
+        "is a deprecated behaviour that will be removed in future.\n"
+        "To set headers to the manifest, use \"inputstream.adaptive.manifest_headers\" property.");
+    m_kodiProps.m_manifestHeaders = m_kodiProps.m_streamHeaders;
+  }
+
+  m_session = std::make_shared<CSession>(m_kodiProps, url, props.GetProfileFolder());
   m_session->SetVideoResolution(m_currentVideoWidth, m_currentVideoHeight, m_currentVideoMaxWidth,
                                 m_currentVideoMaxHeight);
 
@@ -262,7 +277,7 @@ bool CInputStreamAdaptive::OpenStream(int streamid)
   bool needRefetch = false; //Make sure that Kodi fetches changes
   stream->m_isEnabled = true;
 
-  const adaptive::AdaptiveTree::Representation* rep(stream->m_adStream.getRepresentation());
+  adaptive::AdaptiveTree::Representation* rep = stream->m_adStream.getRepresentation();
 
   // If we select a dummy (=inside video) stream, open the video part
   // Dummy streams will be never enabled, they will only enable / activate audio track.
@@ -322,12 +337,27 @@ bool CInputStreamAdaptive::OpenStream(int streamid)
     stream->SetReader(std::make_unique<CTSSampleReader>(
         stream->GetAdByteStream(), stream->m_info.GetStreamType(), streamid, mask));
 
-    if (!stream->GetReader()->Initialize())
+    if (stream->GetReader()->Initialize())
+    {
+      m_session->OnSegmentChanged(&stream->m_adStream);
+    }
+    else if (stream->m_adStream.get_type() == adaptive::AdaptiveTree::AUDIO)
+    {
+      // If TSSampleReader fail, try fallback to ADTS
+      //! @todo: we should have an appropriate file type check
+      //! e.g. with HLS we determine the container type from file extension
+      //! in the url address, but .ts file could have ADTS
+      LOG::LogF(LOGWARNING, "Cannot initialize TS sample reader, fallback to ADTS sample reader");
+      rep->containerType_ = adaptive::AdaptiveTree::CONTAINERTYPE_ADTS;
+
+      stream->GetAdByteStream()->Seek(0); // Seek because bytes are consumed from previous reader
+      stream->SetReader(std::make_unique<CADTSSampleReader>(stream->GetAdByteStream(), streamid));
+    }
+    else
     {
       stream->Disable();
       return false;
     }
-    m_session->OnSegmentChanged(&stream->m_adStream);
   }
   else if (rep->containerType_ == adaptive::AdaptiveTree::CONTAINERTYPE_ADTS)
   {
@@ -462,7 +492,7 @@ DEMUX_PACKET* CInputStreamAdaptive::DemuxRead(void)
         std::memcpy(p->cryptoInfo->kid, pData, 16);
         pData += 16;
         iSize -= (pData - sr->GetSampleData());
-        ReaderCryptoInfo cryptoInfo = sr->GetReaderCryptoInfo();
+        CryptoInfo cryptoInfo = sr->GetReaderCryptoInfo();
         p->cryptoInfo->numSubSamples = numSubSamples;
         p->cryptoInfo->cryptBlocks = cryptoInfo.m_cryptBlocks;
         p->cryptoInfo->skipBlocks = cryptoInfo.m_skipBlocks;
@@ -474,14 +504,15 @@ DEMUX_PACKET* CInputStreamAdaptive::DemuxRead(void)
 
       if (srHaveData)
       {
-        p->dts = static_cast<double>(sr->DTS() + m_session->GetChapterStartTime());
-        p->pts = static_cast<double>(sr->PTS() + m_session->GetChapterStartTime());
+        p->dts = static_cast<double>(sr->DTS());
+        p->pts = static_cast<double>(sr->PTS());
         p->duration = static_cast<double>(sr->GetDuration());
         p->iStreamId = sr->GetStreamId();
         p->iGroupId = 0;
         p->iSize = iSize;
         std::memcpy(p->pData, pData, iSize);
-
+        //! @todo: on Kodi 21, sending of subtitles packet side data is no longer 
+        //! needed and can be removed
         sr->SetDemuxPacketSideData(p, m_session);
       }
 
