@@ -16,6 +16,7 @@
 #include "PRProtectionParser.h"
 #include "kodi/tools/StringUtils.h"
 
+#include <cmath>
 #include <cstring>
 #include <float.h>
 #include <string>
@@ -101,12 +102,12 @@ static uint64_t ParseSegmentTemplate(const char** attr,
   if (!tpl.timescale) // if not specified timescale defaults to seconds
     tpl.timescale = 1;
 
-  if (!URL::IsUrlAbsolute(tpl.media))
+  if (URL::IsUrlRelative(tpl.media))
   {
     tpl.media = URL::Join(baseURL, tpl.media);
   }
 
-  if (!URL::IsUrlAbsolute(tpl.initialization))
+  if (URL::IsUrlRelative(tpl.initialization))
   {
     tpl.initialization = URL::Join(baseURL, tpl.initialization);
   }
@@ -181,7 +182,15 @@ bool ParseContentProtection(const char** attr, DASHTree* dash)
       if (strcmp((const char*)*(attr + 1), "urn:mpeg:dash:mp4protection:2011") == 0)
         mpdFound = true;
       else
-        urnFound = stricmp(dash->supportedKeySystem_.c_str(), (const char*)*(attr + 1)) == 0;
+        urnFound = stricmp(dash->m_supportedKeySystem.c_str(), (const char*)*(attr + 1)) == 0;
+    }
+    else if (StringUtils::EqualsNoCase(*attr, "value"))
+    {
+      std::string_view protectionValue = *(attr + 1);
+      if (protectionValue == "cenc")
+        dash->m_cryptoMode = CryptoMode::AES_CTR;
+      else if (protectionValue == "cbcs")
+        dash->m_cryptoMode = CryptoMode::AES_CBC;
     }
     else if (StringUtils::EndsWith(*attr, "default_KID"))
       defaultKID = (const char*)*(attr + 1);
@@ -363,6 +372,9 @@ static void XMLCALL start(void* data, const char* el, const char** attr)
             {
               dash->current_representation_->flags_ |= DASHTree::Representation::TIMELINE;
               dash->currentNode_ |= MPDNODE_SEGMENTTIMELINE;
+
+              if (dash->m_manifestUpdateParam.empty() && dash->has_timeshift_buffer_)
+                dash->m_manifestUpdateParam = "full";
             }
           }
           else if (dash->currentNode_ & MPDNODE_CONTENTPROTECTION)
@@ -526,8 +538,8 @@ static void XMLCALL start(void* data, const char* el, const char** attr)
             dash->currentNode_ |= MPDNODE_SEGMENTTIMELINE;
             dash->adp_timelined_ = true;
 
-            if (dash->update_parameter_.empty() && dash->has_timeshift_buffer_)
-              dash->update_parameter_ = "full";
+            if (dash->m_manifestUpdateParam.empty() && dash->has_timeshift_buffer_)
+              dash->m_manifestUpdateParam = "full";
           }
         }
         else if (dash->currentNode_ & MPDNODE_SEGMENTDURATIONS)
@@ -999,6 +1011,7 @@ static void XMLCALL start(void* data, const char* el, const char** attr)
       dash->periods_.push_back(dash->current_period_);
       dash->period_timelined_ = false;
       dash->current_period_->start_ = 0;
+      dash->current_period_->sequence_ = dash->current_sequence_++;
 
       for (; *attr;)
       {
@@ -1023,7 +1036,7 @@ static void XMLCALL start(void* data, const char* el, const char** attr)
     const char *mpt(0), *tsbd(0);
 
     dash->firstStartNumber_ = 0;
-
+    dash->current_sequence_ = 0;
     dash->overallSeconds_ = 0;
     dash->stream_start_ = dash->GetNowTime();
     dash->mpd_url_ = dash->base_url_;
@@ -1047,7 +1060,7 @@ static void XMLCALL start(void* data, const char* el, const char** attr)
       else if (strcmp((const char*)*attr, "availabilityStartTime") == 0)
         dash->available_time_ = getTime((const char*)*(attr + 1));
       else if (strcmp((const char*)*attr, "suggestedPresentationDelay") == 0)
-        AddDuration((const char*)*(attr + 1), dash->live_delay_, 1);
+        AddDuration((const char*)*(attr + 1), dash->m_liveDelay, 1);
       else if (strcmp((const char*)*attr, "minimumUpdatePeriod") == 0)
       {
         uint64_t dur(0);
@@ -1126,7 +1139,7 @@ static void XMLCALL end(void* data, const char* el)
                 dash->strXMLText_.erase(dash->strXMLText_.begin());
 
               std::string url{dash->strXMLText_};
-              if (!URL::IsUrlAbsolute(url) && !URL::IsUrlRelative(url))
+              if (URL::IsUrlRelative(url))
                 url = URL::Join(dash->current_adaptationset_->base_url_, url);
 
               dash->current_representation_->base_url_ = url;
@@ -1239,9 +1252,8 @@ static void XMLCALL end(void* data, const char* el)
                 if (countSegs == 0)
                 {
                   countSegs =
-                      static_cast<size_t>(static_cast<double>(overallSeconds) /
-                                          (static_cast<double>(tpl.duration) / tpl.timescale)) +
-                      1;
+                      static_cast<size_t>(std::ceil(static_cast<double>(overallSeconds) /
+                                          (static_cast<double>(tpl.duration) / tpl.timescale)));
                 }
 
                 if (countSegs < 65536)
@@ -1355,7 +1367,9 @@ static void XMLCALL end(void* data, const char* el)
         {
           if (strcmp(el, "BaseURL") == 0)
           {
-            if (URL::IsUrlAbsolute(dash->strXMLText_) || URL::IsUrlRelative(dash->strXMLText_))
+            URL::EnsureEndingBackslash(dash->strXMLText_);
+
+            if (URL::IsUrlAbsolute(dash->strXMLText_))
                 dash->current_adaptationset_->base_url_ = dash->strXMLText_;
             else
               dash->current_adaptationset_->base_url_ =
@@ -1499,7 +1513,10 @@ static void XMLCALL end(void* data, const char* el)
           {
             dash->strXMLText_.erase(dash->strXMLText_.begin());
           }
-          if (URL::IsUrlAbsolute(dash->strXMLText_) || URL::IsUrlRelative(dash->strXMLText_))
+
+          URL::EnsureEndingBackslash(dash->strXMLText_);
+
+          if (URL::IsUrlAbsolute(dash->strXMLText_))
           {
             dash->current_period_->base_url_ = dash->strXMLText_;
           }
@@ -1537,7 +1554,10 @@ static void XMLCALL end(void* data, const char* el)
         {
           dash->strXMLText_.erase(dash->strXMLText_.begin());
         }
-        if (URL::IsUrlAbsolute(dash->strXMLText_) || URL::IsUrlRelative(dash->strXMLText_))
+
+        URL::EnsureEndingBackslash(dash->strXMLText_);
+
+        if (URL::IsUrlAbsolute(dash->strXMLText_))
         {
           dash->mpd_url_ = dash->strXMLText_;
         }
@@ -1594,33 +1614,35 @@ static void XMLCALL end(void* data, const char* el)
   }
 }
 
-adaptive::DASHTree::DASHTree(const DASHTree& left)
-  : AdaptiveTree(left.m_kodiProps, left.m_reprChooser)
+DASHTree::DASHTree(const DASHTree& left) : AdaptiveTree(left)
 {
+  base_time_ = left.base_time_;
+  
+  // Location element should be used on manifest updates
+  location_ = left.location_;
 }
 
 /*----------------------------------------------------------------------
 |   DASHTree
 +---------------------------------------------------------------------*/
-bool DASHTree::open(const std::string& url, const std::string& manifestUpdateParam)
+bool DASHTree::open(const std::string& url)
 {
-  return open(url, manifestUpdateParam, std::map<std::string, std::string>());
+  return open(url, {});
 }
 
-bool DASHTree::open(const std::string& url, const std::string& manifestUpdateParam, std::map<std::string, std::string> additionalHeaders)
+bool DASHTree::open(const std::string& url, std::map<std::string, std::string> addHeaders)
 {
   currentNode_ = 0;
 
-  PrepareManifestUrl(url, manifestUpdateParam);
-  additionalHeaders.insert(m_streamHeaders.begin(), m_streamHeaders.end());
-
   std::stringstream data;
   HTTPRespHeaders respHeaders;
-  if (!download(manifest_url_, additionalHeaders, data, respHeaders))
+  if (!DownloadManifest(url, addHeaders, data, respHeaders))
     return false;
 
+  SaveManifest("", data, url);
+
   effective_url_ = respHeaders.m_effectiveUrl;
-  m_manifestHeaders = respHeaders;
+  m_manifestRespHeaders = respHeaders;
 
   if (!PreparePaths(effective_url_))
     return false;
@@ -1636,11 +1658,20 @@ bool DASHTree::open(const std::string& url, const std::string& manifestUpdatePar
 
   current_period_ = periods_[0];
   SortTree();
-  StartUpdateThread();
 
   return true;
 }
 
+void DASHTree::SetManifestUpdateParam(std::string& manifestUrl, std::string_view param)
+{
+  m_manifestUpdateParam = param;
+  if (m_manifestUpdateParam.empty())
+  {
+    m_manifestUpdateParam = URL::GetParametersFromPlaceholder(manifestUrl, "$START_NUMBER$");
+    manifestUrl.resize(manifestUrl.size() - m_manifestUpdateParam.size());
+  }
+}
+
 bool DASHTree::ParseManifest(const std::string& data)
 {
   strXMLText_.clear();
@@ -1675,8 +1706,7 @@ void DASHTree::RefreshSegments(Period* period,
 {
   if ((type == VIDEO || type == AUDIO))
   {
-    lastUpdated_ = GetTimePointNowTime();
-    RefreshUpdateThread();
+    m_updThread.ResetStartTime();
     RefreshLiveSegments();
   }
 }
@@ -1685,273 +1715,304 @@ void DASHTree::RefreshSegments(Period* period,
 //! @todo: we are updating variables in non-thread safe way
 void DASHTree::RefreshLiveSegments()
 {
-  if (has_timeshift_buffer_ && !update_parameter_.empty())
+  lastUpdated_ = std::chrono::system_clock::now();
+
+  if (m_manifestUpdateParam.empty())
   {
-    size_t numReplace{~(size_t)0};
-    size_t nextStartNumber{~(size_t)0};
+    LOG::LogF(LOGERROR, "Cannot refresh live segments, manifest update param is not set");
+    return;
+  }
 
-    std::string manifestUrlUpd{manifest_url_};
-    bool urlHaveStartNumber{update_parameter_.find("$START_NUMBER$") != std::string::npos};
+  size_t numReplace{~(size_t)0};
+  size_t nextStartNumber{~(size_t)0};
 
-    if (urlHaveStartNumber)
+  std::unique_ptr<DASHTree> updateTree{std::move(Clone())};
+
+  std::string manifestUrlUpd{manifest_url_};
+  bool urlHaveStartNumber{m_manifestUpdateParam.find("$START_NUMBER$") != std::string::npos};
+
+  if (urlHaveStartNumber)
+  {
+    for (auto period : periods_)
     {
-      for (auto period : periods_)
+      if (!period)
+        continue;
+      for (auto adaptSet : period->adaptationSets_)
       {
-        if (!period)
+        if (!adaptSet)
           continue;
-        for (auto adaptSet : period->adaptationSets_)
+        for (auto repr : adaptSet->representations_)
         {
-          if (!adaptSet)
+          if (!repr)
             continue;
-          for (auto repr : adaptSet->representations_)
-          {
-            if (!repr)
-              continue;
 
-            if (repr->startNumber_ + repr->segments_.size() < nextStartNumber)
-              nextStartNumber = repr->startNumber_ + repr->segments_.size();
+          if (repr->startNumber_ + repr->segments_.size() < nextStartNumber)
+            nextStartNumber = repr->startNumber_ + repr->segments_.size();
 
-            size_t replaceable = repr->getCurrentSegmentPos() + 1;
-            if (replaceable == 0) // (~0ULL + 1) == 0
-              replaceable = repr->segments_.size();
+          size_t replaceable = repr->getCurrentSegmentPos() + 1;
+          if (replaceable == 0) // (~0ULL + 1) == 0
+            replaceable = repr->segments_.size();
 
-            if (replaceable < numReplace)
-              numReplace = replaceable;
-          }
+          if (replaceable < numReplace)
+            numReplace = replaceable;
         }
       }
-      LOG::LogF(LOGDEBUG, "Manifest URL start number param set to: %zu (numReplace: %zu)",
-                nextStartNumber, numReplace);
-
-      URL::AppendParameters(manifestUrlUpd, update_parameter_);
-      STRING::ReplaceFirst(manifestUrlUpd, "$START_NUMBER$", std::to_string(nextStartNumber));
     }
+    LOG::LogF(LOGDEBUG, "Manifest URL start number param set to: %zu (numReplace: %zu)",
+              nextStartNumber, numReplace);
 
-    std::unique_ptr<DASHTree> updateTree{std::move(Clone())};
+    // Add the manifest update url parameter with the predefined url parameters
+    std::string updateParam = m_manifestUpdateParam;
 
-    updateTree->base_time_ = base_time_;
-    updateTree->supportedKeySystem_ = supportedKeySystem_;
-    //Location element should be used on updates
-    updateTree->location_ = location_;
+    STRING::ReplaceFirst(updateParam, "$START_NUMBER$", std::to_string(nextStartNumber));
 
-    if (!urlHaveStartNumber)
+    URL::AppendParameters(updateTree->m_manifestParams, updateParam);
+  }
+
+  std::map<std::string, std::string> addHeaders;
+
+  if (!urlHaveStartNumber)
+  {
+    if (!m_manifestRespHeaders.m_etag.empty())
+      addHeaders["If-None-Match"] = "\"" + m_manifestRespHeaders.m_etag + "\"";
+
+    if (!m_manifestRespHeaders.m_lastModified.empty())
+      addHeaders["If-Modified-Since"] = m_manifestRespHeaders.m_lastModified;
+  }
+  //m_manifestParams
+  if (updateTree->open(manifestUrlUpd, addHeaders))
+  {
+    m_manifestRespHeaders = updateTree->m_manifestRespHeaders;
+    location_ = updateTree->location_;
+
+    //Youtube returns last smallest number in case the requested data is not available
+    if (urlHaveStartNumber && updateTree->firstStartNumber_ < nextStartNumber)
+      return;
+
+    for (size_t index{0}; index < updateTree->periods_.size(); index++)
     {
-      if (!m_manifestHeaders.m_etag.empty())
-        updateTree->m_streamHeaders["If-None-Match"] = "\"" + m_manifestHeaders.m_etag + "\"";
-      if (!m_manifestHeaders.m_lastModified.empty())
-        updateTree->m_streamHeaders["If-Modified-Since"] = m_manifestHeaders.m_lastModified;
-    }
+      auto updPeriod = updateTree->periods_[index];
+      if (!updPeriod)
+        continue;
 
-    if (updateTree->open(manifestUrlUpd, ""))
-    {
-      m_manifestHeaders.m_etag = updateTree->m_manifestHeaders.m_etag;
-      m_manifestHeaders.m_lastModified = updateTree->m_manifestHeaders.m_lastModified;
-      location_ = updateTree->location_;
-
-      //Youtube returns last smallest number in case the requested data is not available
-      if (urlHaveStartNumber && updateTree->firstStartNumber_ < nextStartNumber)
-        return;
-
-      for (size_t index{0}; index < updateTree->periods_.size(); index++)
+      Period* period{nullptr};
+      // find matching period based on ID
+      auto itPd = std::find_if(periods_.begin(), periods_.end(),
+                               [&updPeriod](const Period* item)
+                               { return !item->id_.empty() && item->id_ == updPeriod->id_; });
+      if (itPd == periods_.end())
       {
-        if (index == periods_.size()) // Exceeded periods
-          break;
-        auto updPeriod = updateTree->periods_[index];
-        if (!updPeriod)
+        // not found, try matching period based on start
+        itPd = std::find_if(periods_.begin(), periods_.end(),
+                            [&updPeriod](const Period* item)
+                            { return item->start_ && item->start_ == updPeriod->start_; });
+      }
+      // found!
+      if (itPd != periods_.end())
+        period = *itPd;
+      if (!period && updPeriod->id_.empty() && updPeriod->start_ == 0)
+      {
+        // not found, fallback match based on position
+        if (index < periods_.size())
+          period = periods_[index];
+      }
+      // new period, insert it
+      if (!period)
+      {
+        LOG::LogF(LOGDEBUG, "Inserting new Period (id=%s, start=%ld)", updPeriod->id_.c_str(),
+                  updPeriod->start_);
+        updateTree->periods_[index] = nullptr; // remove to prevent delete; we take ownership
+        updPeriod->sequence_ = current_sequence_++;
+        periods_.push_back(updPeriod);
+        continue;
+      }
+
+      for (auto updAdaptationSet : updPeriod->adaptationSets_)
+      {
+        // Locate adaptationset
+        if (!updAdaptationSet)
           continue;
 
-        for (auto updAdaptationSet : updPeriod->adaptationSets_)
+        for (auto adaptationSet : period->adaptationSets_)
         {
-          // Locate adaptationset
-          if (!updAdaptationSet)
+          if (!(adaptationSet->id_ == updAdaptationSet->id_ &&
+                adaptationSet->group_ == updAdaptationSet->group_ &&
+                adaptationSet->type_ == updAdaptationSet->type_ &&
+                adaptationSet->mimeType_ == updAdaptationSet->mimeType_ &&
+                adaptationSet->language_ == updAdaptationSet->language_))
             continue;
 
-          auto& adaptationSets = periods_[index]->adaptationSets_;
-          auto itAs = std::find_if(adaptationSets.begin(), adaptationSets.end(),
-                                   [&updAdaptationSet](const AdaptationSet* item)
-                                   {
-                                     return item->id_ == updAdaptationSet->id_ &&
-                                            item->group_ == updAdaptationSet->group_ &&
-                                            item->type_ == updAdaptationSet->type_ &&
-                                            item->mimeType_ == updAdaptationSet->mimeType_ &&
-                                            item->language_ == updAdaptationSet->language_;
-                                   });
-          if (itAs != adaptationSets.end()) // Found adaptationset
+          for (auto updRepr : updAdaptationSet->representations_)
           {
-            auto adaptationSet = *itAs;
+            // Locate representation
+            auto itR = std::find_if(
+                adaptationSet->representations_.begin(), adaptationSet->representations_.end(),
+                [&updRepr](const Representation* item) { return item->id == updRepr->id; });
 
-            for (auto updRepr : updAdaptationSet->representations_)
+            if (!updRepr->segments_.Get(0))
             {
-              // Locate representation
-              auto itR = std::find_if(
-                  adaptationSet->representations_.begin(), adaptationSet->representations_.end(),
-                  [&updRepr](const Representation* item) { return item->id == updRepr->id; });
+              LOG::LogF(LOGERROR,
+                        "Segment at position 0 not found from (update) representation id: %s",
+                        updRepr->id.c_str());
+              return;
+            }
 
-              if (!updRepr->segments_.Get(0))
+            if (itR != adaptationSet->representations_.end()) // Found representation
+            {
+              auto repr = *itR;
+
+              if (!repr->segments_.empty())
               {
-                LOG::LogF(LOGERROR,
-                          "Segment at position 0 not found from (update) representation id: %s",
-                          updRepr->id.c_str());
-                return;
-              }
-
-              if (itR != adaptationSet->representations_.end()) // Found representation
-              {
-                auto repr = *itR;
-
-                if (!repr->segments_.empty())
+                if (urlHaveStartNumber) // Partitial update
                 {
-                  if (urlHaveStartNumber) // Partitial update
+                  // Insert new segments
+                  uint64_t ptsOffset = repr->nextPts_ - updRepr->segments_.Get(0)->startPTS_;
+                  size_t currentPos = repr->getCurrentSegmentPos();
+                  size_t repFreeSegments{numReplace};
+
+                  auto updSegmentIt(updRepr->segments_.data.begin());
+                  for (; updSegmentIt != updRepr->segments_.data.end() && repFreeSegments != 0;
+                       updSegmentIt++)
                   {
-                    // Insert new segments
-                    uint64_t ptsOffset = repr->nextPts_ - updRepr->segments_.Get(0)->startPTS_;
-                    size_t currentPos = repr->getCurrentSegmentPos();
-                    size_t repFreeSegments{numReplace};
-
-                    auto updSegmentIt(updRepr->segments_.data.begin());
-                    for (; updSegmentIt != updRepr->segments_.data.end() && repFreeSegments != 0;
-                         updSegmentIt++)
+                    LOG::LogF(LOGDEBUG, "Insert representation (id: %s url: %s)",
+                              updRepr->id.c_str(), updSegmentIt->url.c_str());
+                    if (repr->flags_ & Representation::URLSEGMENTS)
                     {
-                      LOG::LogF(LOGDEBUG, "Insert representation (id: %s url: %s)",
-                                updRepr->id.c_str(), updSegmentIt->url.c_str());
-                      if (repr->flags_ & Representation::URLSEGMENTS)
+                      Segment* segment{repr->segments_.Get(0)};
+                      if (!segment)
                       {
-                        Segment* segment{repr->segments_.Get(0)};
-                        if (!segment)
-                        {
-                          LOG::LogF(LOGERROR,
-                                    "Segment at position 0 not found from representation id: %s",
-                                    repr->id.c_str());
-                          return;
-                        }
-                        segment->url.clear();
+                        LOG::LogF(LOGERROR,
+                                  "Segment at position 0 not found from representation id: %s",
+                                  repr->id.c_str());
+                        return;
                       }
-                      updSegmentIt->startPTS_ += ptsOffset;
-                      repr->segments_.insert(*updSegmentIt);
-                      if (repr->flags_ & Representation::URLSEGMENTS)
-                        updSegmentIt->url.clear();
-                      repr->startNumber_++;
-                      repFreeSegments--;
+                      segment->url.clear();
                     }
-
-                    //We have renewed the current segment
-                    if (!repFreeSegments && numReplace == currentPos + 1)
-                      repr->current_segment_ = nullptr;
-
-                    if ((repr->flags_ & Representation::WAITFORSEGMENT) &&
-                        repr->get_next_segment(repr->current_segment_))
-                    {
-                      repr->flags_ &= ~Representation::WAITFORSEGMENT;
-                      LOG::LogF(LOGDEBUG, "End WaitForSegment stream %s", repr->id.c_str());
-                    }
-
-                    if (updSegmentIt == updRepr->segments_.data.end())
-                      repr->nextPts_ += updRepr->nextPts_;
-                    else
-                      repr->nextPts_ += updSegmentIt->startPTS_;
+                    updSegmentIt->startPTS_ += ptsOffset;
+                    repr->segments_.insert(*updSegmentIt);
+                    if (repr->flags_ & Representation::URLSEGMENTS)
+                      updSegmentIt->url.clear();
+                    repr->startNumber_++;
+                    repFreeSegments--;
                   }
-                  else if (updRepr->startNumber_ <= 1) // Full update, be careful with startnumbers!
+
+                  //We have renewed the current segment
+                  if (!repFreeSegments && numReplace == currentPos + 1)
+                    repr->current_segment_ = nullptr;
+
+                  if ((repr->flags_ & Representation::WAITFORSEGMENT) &&
+                      repr->get_next_segment(repr->current_segment_))
                   {
-                    //TODO: check if first element or size differs
-                    uint64_t segmentId(repr->getCurrentSegmentNumber());
-                    if (repr->flags_ & DASHTree::Representation::TIMELINE)
+                    repr->flags_ &= ~Representation::WAITFORSEGMENT;
+                    LOG::LogF(LOGDEBUG, "End WaitForSegment stream %s", repr->id.c_str());
+                  }
+
+                  if (updSegmentIt == updRepr->segments_.data.end())
+                    repr->nextPts_ += updRepr->nextPts_;
+                  else
+                    repr->nextPts_ += updSegmentIt->startPTS_;
+                }
+                else if (updRepr->startNumber_ <= 1) // Full update, be careful with startnumbers!
+                {
+                  //TODO: check if first element or size differs
+                  uint64_t segmentId(repr->getCurrentSegmentNumber());
+                  if (repr->flags_ & DASHTree::Representation::TIMELINE)
+                  {
+                    uint64_t search_pts = updRepr->segments_.Get(0)->range_begin_;
+                    uint64_t misaligned = 0;
+                    for (const auto& segment : repr->segments_.data)
                     {
-                      uint64_t search_pts = updRepr->segments_.Get(0)->range_begin_;
-                      uint64_t misaligned = 0;
-                      for (const auto& segment : repr->segments_.data)
+                      if (misaligned)
                       {
-                        if (misaligned)
-                        {
-                          uint64_t ptsDiff = segment.range_begin_ - (&segment - 1)->range_begin_;
-                          // our misalignment is small ( < 2%), let's decrement the start number
-                          if (misaligned < (ptsDiff * 2 / 100))
-                            --repr->startNumber_;
-                          break;
-                        }
-                        if (segment.range_begin_ == search_pts)
-                          break;
-                        else if (segment.range_begin_ > search_pts)
-                        {
-                          if (&repr->segments_.data.front() == &segment)
-                          {
-                            --repr->startNumber_;
-                            break;
-                          }
-                          misaligned = search_pts - (&segment - 1)->range_begin_;
-                        }
-                        else
-                          ++repr->startNumber_;
+                        uint64_t ptsDiff = segment.range_begin_ - (&segment - 1)->range_begin_;
+                        // our misalignment is small ( < 2%), let's decrement the start number
+                        if (misaligned < (ptsDiff * 2 / 100))
+                          --repr->startNumber_;
+                        break;
                       }
-                    }
-                    else if (updRepr->segments_.Get(0)->startPTS_ ==
-                             repr->segments_.Get(0)->startPTS_)
-                    {
-                      uint64_t search_re = updRepr->segments_.Get(0)->range_end_;
-                      for (const auto& segment : repr->segments_.data)
+                      if (segment.range_begin_ == search_pts)
+                        break;
+                      else if (segment.range_begin_ > search_pts)
                       {
-                        if (segment.range_end_ >= search_re)
+                        if (&repr->segments_.data.front() == &segment)
+                        {
+                          --repr->startNumber_;
                           break;
+                        }
+                        misaligned = search_pts - (&segment - 1)->range_begin_;
+                      }
+                      else
                         ++repr->startNumber_;
-                      }
                     }
-                    else
-                    {
-                      uint64_t search_pts = updRepr->segments_.Get(0)->startPTS_;
-                      for (const auto& segment : repr->segments_.data)
-                      {
-                        if (segment.startPTS_ >= search_pts)
-                          break;
-                        ++repr->startNumber_;
-                      }
-                    }
-
-                    updRepr->segments_.swap(repr->segments_);
-                    if (!~segmentId || segmentId < repr->startNumber_)
-                      repr->current_segment_ = nullptr;
-                    else
-                    {
-                      if (segmentId >= repr->startNumber_ + repr->segments_.size())
-                        segmentId = repr->startNumber_ + repr->segments_.size() - 1;
-                      repr->current_segment_ =
-                          repr->get_segment(static_cast<size_t>(segmentId - repr->startNumber_));
-                    }
-
-                    if ((repr->flags_ & Representation::WAITFORSEGMENT) &&
-                        repr->get_next_segment(repr->current_segment_))
-                      repr->flags_ &= ~Representation::WAITFORSEGMENT;
-
-                    LOG::LogF(
-                        LOGDEBUG,
-                        "Full update without start number (repr. id: %s current start: %u)",
-                        updRepr->id.c_str(), repr->startNumber_);
-                    overallSeconds_ = updateTree->overallSeconds_;
                   }
-                  else if (updRepr->startNumber_ > repr->startNumber_ ||
-                           (updRepr->startNumber_ == repr->startNumber_ &&
-                            updRepr->segments_.size() > repr->segments_.size()))
+                  else if (updRepr->segments_.Get(0)->startPTS_ ==
+                           repr->segments_.Get(0)->startPTS_)
                   {
-                    size_t segmentId(repr->getCurrentSegmentNumber());
-                    updRepr->segments_.swap(repr->segments_);
-                    repr->startNumber_ = updRepr->startNumber_;
-                    if (!~segmentId || segmentId < repr->startNumber_)
-                      repr->current_segment_ = nullptr;
-                    else
+                    uint64_t search_re = updRepr->segments_.Get(0)->range_end_;
+                    for (const auto& segment : repr->segments_.data)
                     {
-                      if (segmentId >= repr->startNumber_ + repr->segments_.size())
-                        segmentId = repr->startNumber_ + repr->segments_.size() - 1;
-                      repr->current_segment_ =
-                          repr->get_segment(static_cast<size_t>(segmentId - repr->startNumber_));
+                      if (segment.range_end_ >= search_re)
+                        break;
+                      ++repr->startNumber_;
                     }
-
-                    if ((repr->flags_ & Representation::WAITFORSEGMENT) &&
-                        repr->get_next_segment(repr->current_segment_))
-                    {
-                      repr->flags_ &= ~Representation::WAITFORSEGMENT;
-                    }
-                    LOG::LogF(LOGDEBUG,
-                              "Full update with start number (repr. id: %s current start:%u)",
-                              updRepr->id.c_str(), repr->startNumber_);
                   }
+                  else
+                  {
+                    uint64_t search_pts = updRepr->segments_.Get(0)->startPTS_;
+                    for (const auto& segment : repr->segments_.data)
+                    {
+                      if (segment.startPTS_ >= search_pts)
+                        break;
+                      ++repr->startNumber_;
+                    }
+                  }
+
+                  updRepr->segments_.swap(repr->segments_);
+                  if (!~segmentId || segmentId < repr->startNumber_)
+                    repr->current_segment_ = nullptr;
+                  else
+                  {
+                    if (segmentId >= repr->startNumber_ + repr->segments_.size())
+                      segmentId = repr->startNumber_ + repr->segments_.size() - 1;
+                    repr->current_segment_ =
+                        repr->get_segment(static_cast<size_t>(segmentId - repr->startNumber_));
+                  }
+
+                  if ((repr->flags_ & Representation::WAITFORSEGMENT) &&
+                      repr->get_next_segment(repr->current_segment_))
+                    repr->flags_ &= ~Representation::WAITFORSEGMENT;
+
+                  LOG::LogF(LOGDEBUG,
+                            "Full update without start number (repr. id: %s current start: %u)",
+                            updRepr->id.c_str(), repr->startNumber_);
+                  overallSeconds_ = updateTree->overallSeconds_;
+                }
+                else if (updRepr->startNumber_ > repr->startNumber_ ||
+                         (updRepr->startNumber_ == repr->startNumber_ &&
+                          updRepr->segments_.size() > repr->segments_.size()))
+                {
+                  size_t segmentId(repr->getCurrentSegmentNumber());
+                  updRepr->segments_.swap(repr->segments_);
+                  repr->startNumber_ = updRepr->startNumber_;
+                  if (!~segmentId || segmentId < repr->startNumber_)
+                    repr->current_segment_ = nullptr;
+                  else
+                  {
+                    if (segmentId >= repr->startNumber_ + repr->segments_.size())
+                      segmentId = repr->startNumber_ + repr->segments_.size() - 1;
+                    repr->current_segment_ =
+                        repr->get_segment(static_cast<size_t>(segmentId - repr->startNumber_));
+                  }
+
+                  if ((repr->flags_ & Representation::WAITFORSEGMENT) &&
+                      repr->get_next_segment(repr->current_segment_))
+                  {
+                    repr->flags_ &= ~Representation::WAITFORSEGMENT;
+                  }
+                  LOG::LogF(LOGDEBUG,
+                            "Full update with start number (repr. id: %s current start:%u)",
+                            updRepr->id.c_str(), repr->startNumber_);
+                  overallSeconds_ = updateTree->overallSeconds_;
                 }
               }
             }
