@@ -74,8 +74,19 @@ static std::string getVideoCodec(const std::string& codecs)
     return "av01";
   else if (codecs.find("av1") != std::string::npos)
     return "av1";
-  else
-    return "";
+  else if (codecs.find("vp9") != std::string::npos || codecs.find("vp09") != std::string::npos)
+  {
+    std::vector<std::string> list = StringUtils::Split(codecs, ",");
+    for (const std::string& name : list)
+    {
+      if (name.find("vp9") != std::string::npos || name.find("vp09") != std::string::npos)
+      {
+        return name;
+      }
+    }
+  }
+
+  return "";
 }
 
 static std::string getAudioCodec(const std::string& codecs)
@@ -88,15 +99,20 @@ static std::string getAudioCodec(const std::string& codecs)
     return "aac";
 }
 
-HLSTree::HLSTree(const UTILS::PROPERTIES::KodiProperties& kodiProps,
-                 CHOOSER::IRepresentationChooser* reprChooser)
-  : AdaptiveTree(kodiProps, reprChooser)
+HLSTree::HLSTree(CHOOSER::IRepresentationChooser* reprChooser)
+  : AdaptiveTree(reprChooser)
 {
-  m_decrypter = std::make_unique<AESDecrypter>(kodiProps.m_licenseKey);
 };
 
-HLSTree::HLSTree(const HLSTree& left) : AdaptiveTree(left.m_kodiProps, left.m_reprChooser)
+HLSTree::HLSTree(const HLSTree& left) : AdaptiveTree(left)
 {
+  m_decrypter = std::make_unique<AESDecrypter>(left.m_decrypter->getLicenseKey());
+}
+
+void HLSTree::Configure(const UTILS::PROPERTIES::KodiProperties& kodiProps)
+{
+  AdaptiveTree::Configure(kodiProps);
+  m_decrypter = std::make_unique<AESDecrypter>(kodiProps.m_licenseKey);
 }
 
 HLSTree::~HLSTree()
@@ -117,7 +133,7 @@ int HLSTree::processEncryption(std::string baseUrl, std::map<std::string, std::s
   if (map["METHOD"] == "AES-128" && !map["URI"].empty())
   {
     current_pssh_ = map["URI"];
-    if (!URL::IsUrlRelative(current_pssh_) && !URL::IsUrlAbsolute(current_pssh_))
+    if (URL::IsUrlRelative(current_pssh_))
       current_pssh_ = URL::Join(baseUrl, current_pssh_);
 
     current_iv_ = m_decrypter->convertIV(map["IV"]);
@@ -170,21 +186,20 @@ int HLSTree::processEncryption(std::string baseUrl, std::map<std::string, std::s
   return ENCRYPTIONTYPE_UNKNOWN;
 }
 
-bool HLSTree::open(const std::string& url, const std::string& manifestUpdateParam)
+bool HLSTree::open(const std::string& url)
 {
-  return open(url, manifestUpdateParam, std::map<std::string, std::string>());
+  return open(url, std::map<std::string, std::string>());
 }
 
-bool HLSTree::open(const std::string& url, const std::string& manifestUpdateParam, std::map<std::string, std::string> additionalHeaders)
+bool HLSTree::open(const std::string& url, std::map<std::string, std::string> addHeaders)
 {
-  PrepareManifestUrl(url, manifestUpdateParam);
-  additionalHeaders.insert(m_streamHeaders.begin(), m_streamHeaders.end());
-
   std::stringstream data;
   HTTPRespHeaders respHeaders;
-  if (!download(manifest_url_, additionalHeaders, data, respHeaders))
+  if (!DownloadManifest(url, addHeaders, data, respHeaders))
     return false;
 
+  SaveManifest(nullptr, data, url);
+
   effective_url_ = respHeaders.m_effectiveUrl;
 
   if (!PreparePaths(effective_url_))
@@ -201,12 +216,6 @@ bool HLSTree::open(const std::string& url, const std::string& manifestUpdatePara
 
 bool HLSTree::ParseManifest(std::stringstream& stream)
 {
-#if FILEDEBUG
-  FILE* f = fopen("inputstream_adaptive_master.m3u8", "w");
-  fwrite(m_stream.str().data(), 1, m_stream.str().size(), f);
-  fclose(f);
-#endif
-
   std::string line;
   bool startCodeFound = false;
 
@@ -417,10 +426,31 @@ bool HLSTree::ParseManifest(std::stringstream& stream)
   }
   // Set Live as default
   has_timeshift_buffer_ = true;
-  update_parameter_ = "full";
+  m_manifestUpdateParam = "full";
   return true;
 }
 
+void HLSTree::SaveManifest(AdaptationSet* adp, const std::stringstream& data, std::string_view info)
+{
+  if (m_pathSaveManifest.empty())
+    return;
+
+  std::string fileNameSuffix = "master";
+  if (adp)
+  {
+    if (adp->type_ == VIDEO)
+      fileNameSuffix = "child-video";
+    else if (adp->type_ == AUDIO)
+      fileNameSuffix = "child-audio";
+    else if (adp->type_ == SUBTITLE)
+      fileNameSuffix = "child-subtitle";
+    else
+      fileNameSuffix = "child";
+  }
+
+  AdaptiveTree::SaveManifest(fileNameSuffix, data, info);
+}
+
 HLSTree::PREPARE_RESULT HLSTree::prepareRepresentation(Period* period,
                                                        AdaptationSet* adp,
                                                        Representation* rep,
@@ -447,13 +477,10 @@ HLSTree::PREPARE_RESULT HLSTree::prepareRepresentation(Period* period,
 
     if (rep->flags_ & Representation::DOWNLOADED) {
     }
-    else if (download(rep->source_url_, m_streamHeaders, streamData, respHeaders))
+    else if (DownloadManifest(rep->source_url_, {}, streamData, respHeaders))
     {
-#if FILEDEBUG
-      FILE* f = fopen("inputstream_adaptive_sub.m3u8", "w");
-      fwrite(m_stream.str().data(), 1, m_stream.str().size(), f);
-      fclose(f);
-#endif
+      SaveManifest(adp, streamData, rep->source_url_);
+
       bool byteRange(false);
       bool segmentInitialization(false);
       bool hasMap(false);
@@ -530,8 +557,20 @@ HLSTree::PREPARE_RESULT HLSTree::prepareRepresentation(Period* period,
                 rep->containerType_ = CONTAINERTYPE_TEXT;
               else
               {
-                rep->containerType_ = CONTAINERTYPE_INVALID;
-                continue;
+                if (adp->type_ == VIDEO)
+                {
+                  // Streams that have a media url encoded as a parameter of the url itself
+                  // cannot be detected in safe way, so we try fallback to .ts
+                  // e.g. https://cdn-prod.tv/beacon?streamId=1&rp=https%3A%2F%2Ftest.com%2F167037ac3%2Findex_4_0.ts&sessionId=abc&assetId=OD
+                  rep->containerType_ = CONTAINERTYPE_TS;
+                  LOG::LogF(LOGDEBUG, "Cannot detect container type from media url, fallback to TS");
+                }
+                else
+                {
+                  rep->containerType_ = CONTAINERTYPE_INVALID;
+                  LOG::LogF(LOGDEBUG, "Cannot detect container type from media url");
+                  continue;
+                }
               }
             }
             else
@@ -544,7 +583,7 @@ HLSTree::PREPARE_RESULT HLSTree::prepareRepresentation(Period* period,
           if (!byteRange || rep->url_.empty())
           {
             std::string url;
-            if (!URL::IsUrlRelative(line) && !URL::IsUrlAbsolute(line))
+            if (URL::IsUrlRelative(line))
               url = URL::Join(base_url, line);
             else
               url = line;
@@ -580,8 +619,8 @@ HLSTree::PREPARE_RESULT HLSTree::prepareRepresentation(Period* period,
         else if (line.compare(0, 22, "#EXT-X-TARGETDURATION:") == 0)
         {
           uint32_t newInterval = atoi(line.c_str() + 22) * 1500;
-          if (newInterval < updateInterval_)
-            updateInterval_ = newInterval;
+          if (newInterval < m_updateInterval)
+            m_updateInterval = newInterval;
         }
         else if (line.compare(0, 30, "#EXT-X-DISCONTINUITY-SEQUENCE:") == 0)
         {
@@ -658,6 +697,7 @@ HLSTree::PREPARE_RESULT HLSTree::prepareRepresentation(Period* period,
           segment.range_end_ = 0;
           segment.startPTS_ = ~0ULL;
           segment.pssh_set_ = 0;
+          pts = 0;
 
           if (currentEncryptionType == ENCRYPTIONTYPE_WIDEVINE)
           {
@@ -707,7 +747,7 @@ HLSTree::PREPARE_RESULT HLSTree::prepareRepresentation(Period* period,
           {
             std::string uri = map["URI"];
 
-            if (!URL::IsUrlRelative(uri) && !URL::IsUrlAbsolute(uri))
+            if (URL::IsUrlRelative(uri))
               map_url = URL::Join(base_url, uri);
             else
               map_url = uri;
@@ -823,7 +863,7 @@ void HLSTree::OnDataArrived(uint64_t segNum,
 {
   if (psshSet && current_period_->encryptionState_ != ENCRYTIONSTATE_SUPPORTED)
   {
-    std::lock_guard<std::mutex> lck(treeMutex_);
+    std::lock_guard<TreeUpdateThread> lckUpdTree(GetTreeUpdMutex());
 
     Period::PSSH& pssh(current_period_->psshSets_[psshSet]);
     //Encrypted media, decrypt it
@@ -902,7 +942,8 @@ void HLSTree::RefreshSegments(Period* period,
   {
     if (rep->flags_ & Representation::INCLUDEDSTREAM)
       return;
-    RefreshUpdateThread();
+    
+    m_updThread.ResetStartTime();
     prepareRepresentation(period, adp, rep, true);
   }
 }
@@ -911,6 +952,8 @@ void HLSTree::RefreshSegments(Period* period,
 //! @todo: we are updating variables in non-thread safe way
 void HLSTree::RefreshLiveSegments()
 {
+  lastUpdated_ = std::chrono::system_clock::now();
+
   if (m_refreshPlayList)
   {
     std::vector<std::tuple<AdaptationSet*, Representation*>> refresh_list;
