@@ -29,8 +29,6 @@
 #error "WIDEVINECDMFILENAME must be set"
 #endif
 
-//#define LOCLICENSE
-
 using namespace SSD;
 using namespace UTILS;
 using namespace kodi::tools;
@@ -173,8 +171,12 @@ public:
   void AddSessionKey(const uint8_t *data, size_t data_size, uint32_t status);
   bool HasKeyId(const uint8_t *keyid);
 
-  virtual AP4_Result SetFragmentInfo(AP4_UI32 pool_id, const AP4_UI08 *key, const AP4_UI08 nal_length_size,
-    AP4_DataBuffer &annexb_sps_pps, AP4_UI32 flags)override;
+  virtual AP4_Result SetFragmentInfo(AP4_UI32 pool_id,
+                                     const AP4_UI08* key,
+                                     const AP4_UI08 nal_length_size,
+                                     AP4_DataBuffer& annexb_sps_pps,
+                                     AP4_UI32 flags,
+                                     CryptoInfo cryptoInfo) override;
   virtual AP4_UI32 AddPool() override;
   virtual void RemovePool(AP4_UI32 poolid) override;
 
@@ -200,7 +202,6 @@ public:
   SSD_DECODE_RETVAL VideoFrameDataToPicture(void* hostInstance, SSD_PICTURE *picture);
   void ResetVideo();
 
-  void SetEncryptionMode(CryptoMode encryptionMode) override;
   void SetDefaultKeyId(std::string_view keyId) override;
   void AddKeyId(std::string_view keyId) override;
 
@@ -232,6 +233,7 @@ private:
     AP4_UI08 nal_length_size_;
     AP4_UI16 decrypter_flags_;
     AP4_DataBuffer annexb_sps_pps_;
+    CryptoInfo m_cryptoInfo;
   };
   std::vector<FINFO> fragment_pool_;
   void LogDecryptError(const cdm::Status status, const AP4_UI08* key);
@@ -436,17 +438,16 @@ WV_CencSingleSampleDecrypter::WV_CencSingleSampleDecrypter(WV_DRM& drm,
 
   drm_.insertssd(this);
 
-#ifdef LOCLICENSE
-  std::string strDbg = GLOBAL::Host->GetProfilePath();
-  strDbg += "EDEF8BA9-79D6-4ACE-A3C8-27DCD51D21ED.init";
-  FILE*f = fopen(strDbg.c_str(), "wb");
-  if (f) {
-    fwrite(pssh.GetData(), 1, pssh.GetDataSize(), f);
-    fclose(f);
+  if (GLOBAL::Host->IsDebugSaveLicense())
+  {
+    //! @todo: with ssd_wv refactor the path must be combined with
+    //!        UTILS::FILESYS::PathCombine
+    std::string debugFilePath = GLOBAL::Host->GetProfilePath();
+    debugFilePath += "EDEF8BA9-79D6-4ACE-A3C8-27DCD51D21ED.init";
+
+    std::string data{reinterpret_cast<const char*>(pssh.GetData()), pssh.GetDataSize()};
+    SSD_UTILS::SaveFile(debugFilePath, data);
   }
-  else
-    LOG::LogF(SSDDEBUG, "Could not open debug file for writing (init)!");
-#endif
 
   if (memcmp(pssh.GetData() + 4, "pssh", 4) != 0)
   {
@@ -497,13 +498,17 @@ void WV_CencSingleSampleDecrypter::GetCapabilities(const uint8_t* key, uint32_t
 {
   caps = { 0, hdcp_version_, hdcp_limit_ };
 
-  if (session_.empty())
+  if (session_.empty()) {
+    LOG::LogF(SSDDEBUG, "Session empty");
     return;
+  }
 
   caps.flags = SSD_DECRYPTER::SSD_CAPS::SSD_SUPPORTS_DECODING;
 
-  if (keys_.empty())
+  if (keys_.empty()) {
+    LOG::LogF(SSDDEBUG, "Keys empty");
     return;
+  }
 
   if (!caps.hdcpLimit)
     caps.hdcpLimit = resolution_limit_;
@@ -524,10 +529,11 @@ void WV_CencSingleSampleDecrypter::GetCapabilities(const uint8_t* key, uint32_t
       break;
     }
     */
-  if (caps.flags == SSD_DECRYPTER::SSD_CAPS::SSD_SUPPORTS_DECODING)
+  if ((caps.flags & SSD_DECRYPTER::SSD_CAPS::SSD_SUPPORTS_DECODING) != 0)
   {
     AP4_UI32 poolid(AddPool());
     fragment_pool_[poolid].key_ = key ? key : reinterpret_cast<const uint8_t*>(keys_.front().keyid.data());
+    fragment_pool_[poolid].m_cryptoInfo.m_mode = m_EncryptionMode;
 
     AP4_DataBuffer in, out;
     AP4_UI32 encb[2] = { 1,1 };
@@ -537,34 +543,31 @@ void WV_CencSingleSampleDecrypter::GetCapabilities(const uint8_t* key, uint32_t
     in.SetBuffer(vf,12);
     in.SetDataSize(12);
     try {
-      if (DecryptSampleData(poolid, in, out, iv, 2, clearb, encb) != AP4_SUCCESS)
+      encb[0] = 12;
+      clearb[0] = 0;
+      if (DecryptSampleData(poolid, in, out, iv, 1, clearb, encb) != AP4_SUCCESS)
       {
-        encb[0] = 12;
-        clearb[0] = 0;
-        if (DecryptSampleData(poolid, in, out, iv, 1, clearb, encb) != AP4_SUCCESS)
-        {
-          if (media == SSD_DECRYPTER::SSD_CAPS::SSD_MEDIA_VIDEO)
-            caps.flags |= (SSD_DECRYPTER::SSD_CAPS::SSD_SECURE_PATH | SSD_DECRYPTER::SSD_CAPS::SSD_ANNEXB_REQUIRED);
-          else
-            caps.flags = SSD_DECRYPTER::SSD_CAPS::SSD_INVALID;
-        }
+        LOG::LogF(SSDDEBUG, "Single decrypt failed, secure path only");
+        if (media == SSD_DECRYPTER::SSD_CAPS::SSD_MEDIA_VIDEO)
+          caps.flags |= (SSD_DECRYPTER::SSD_CAPS::SSD_SECURE_PATH | SSD_DECRYPTER::SSD_CAPS::SSD_ANNEXB_REQUIRED);
         else
-        {
-          caps.flags |= SSD_DECRYPTER::SSD_CAPS::SSD_SINGLE_DECRYPT;
-          caps.hdcpVersion = 99;
-          caps.hdcpLimit = resolution_limit_;
-        }
+          caps.flags = SSD_DECRYPTER::SSD_CAPS::SSD_INVALID;
       }
       else
       {
+        LOG::LogF(SSDDEBUG, "Single decrypt possible");
+        caps.flags |= SSD_DECRYPTER::SSD_CAPS::SSD_SINGLE_DECRYPT;
         caps.hdcpVersion = 99;
         caps.hdcpLimit = resolution_limit_;
       }
     }
-    catch (...) {
+    catch (const std::exception& e) {
+      LOG::LogF(SSDDEBUG, "Decrypt error, assuming secure path: %s", e.what());
       caps.flags |= (SSD_DECRYPTER::SSD_CAPS::SSD_SECURE_PATH | SSD_DECRYPTER::SSD_CAPS::SSD_ANNEXB_REQUIRED);
     }
     RemovePool(poolid);
+  } else {
+    LOG::LogF(SSDDEBUG, "Decoding not supported");
   }
 }
 
@@ -611,17 +614,16 @@ bool WV_CencSingleSampleDecrypter::SendSessionMessage()
     return false;
   }
 
-#ifdef LOCLICENSE
-  std::string strDbg = GLOBAL::Host->GetProfilePath();
-  strDbg += "EDEF8BA9-79D6-4ACE-A3C8-27DCD51D21ED.challenge";
-  FILE*f = fopen(strDbg.c_str(), "wb");
-  if (f) {
-    fwrite(challenge_.GetData(), 1, challenge_.GetDataSize(), f);
-    fclose(f);
+  if (GLOBAL::Host->IsDebugSaveLicense())
+  {
+    //! @todo: with ssd_wv refactor the path must be combined with
+    //!        UTILS::FILESYS::PathCombine
+    std::string debugFilePath = GLOBAL::Host->GetProfilePath();
+    debugFilePath += "EDEF8BA9-79D6-4ACE-A3C8-27DCD51D21ED.challenge";
+
+    std::string data{reinterpret_cast<const char*>(challenge_.GetData()), challenge_.GetDataSize()};
+    SSD_UTILS::SaveFile(debugFilePath, data);
   }
-  else
-    LOG::LogF(SSDDEBUG, "Could not open debug file for writing (challenge)!");
-#endif
 
   //Process placeholder in GET String
   std::string::size_type insPos(blocks[0].find("{SSM}"));
@@ -666,14 +668,18 @@ bool WV_CencSingleSampleDecrypter::SendSessionMessage()
   for (std::string& headerStr : headers)
   {
     std::vector<std::string> header{StringUtils::Split(headerStr, '=')};
-    StringUtils::Trim(header[0]);
-    std::string value;
-    if (header.size() > 1)
+    if (!header.empty())
     {
-      StringUtils::Trim(header[1]);
-      value = STRING::URLDecode(header[1]);
+      StringUtils::Trim(header[0]);
+      std::string value;
+      if (header.size() > 1)
+      {
+        StringUtils::Trim(header[1]);
+        value = STRING::URLDecode(header[1]);
+      }
+      GLOBAL::Host->CURLAddOption(file, SSD_HOST::OPTION_PROTOCOL, header[0].c_str(),
+                                  value.c_str());
     }
-    GLOBAL::Host->CURLAddOption(file, SSD_HOST::OPTION_PROTOCOL, header[0].c_str(), value.c_str());
   }
 
   //Process body
@@ -833,17 +839,15 @@ bool WV_CencSingleSampleDecrypter::SendSessionMessage()
     goto SSMFAIL;
   }
 
-#ifdef LOCLICENSE
-  strDbg = GLOBAL::Host->GetProfilePath();
-  strDbg += "EDEF8BA9-79D6-4ACE-A3C8-27DCD51D21ED.response";
-  f = fopen(strDbg.c_str(), "wb");
-  if (f) {
-    fwrite(response.c_str(), 1, response.size(), f);
-    fclose(f);
+  if (GLOBAL::Host->IsDebugSaveLicense())
+  {
+    //! @todo: with ssd_wv refactor the path must be combined with
+    //!        UTILS::FILESYS::PathCombine
+    std::string debugFilePath = GLOBAL::Host->GetProfilePath();
+    debugFilePath += "EDEF8BA9-79D6-4ACE-A3C8-27DCD51D21ED.response";
+
+    SSD_UTILS::SaveFile(debugFilePath, response);
   }
-  else
-    LOG::LogF(SSDDEBUG, "Could not open debug file for writing (response)!");
-#endif
 
   if (serverCertRequest && contentType.find("application/octet-stream") == std::string::npos)
     serverCertRequest = false;
@@ -997,7 +1001,12 @@ bool WV_CencSingleSampleDecrypter::HasKeyId(const uint8_t *keyid)
   return false;
 }
 
-AP4_Result WV_CencSingleSampleDecrypter::SetFragmentInfo(AP4_UI32 pool_id, const AP4_UI08 *key, const AP4_UI08 nal_length_size, AP4_DataBuffer &annexb_sps_pps, AP4_UI32 flags)
+AP4_Result WV_CencSingleSampleDecrypter::SetFragmentInfo(AP4_UI32 pool_id,
+                                                         const AP4_UI08* key,
+                                                         const AP4_UI08 nal_length_size,
+                                                         AP4_DataBuffer& annexb_sps_pps,
+                                                         AP4_UI32 flags,
+                                                         CryptoInfo cryptoInfo)
 {
   if (pool_id >= fragment_pool_.size())
     return AP4_ERROR_OUT_OF_RANGE;
@@ -1006,6 +1015,7 @@ AP4_Result WV_CencSingleSampleDecrypter::SetFragmentInfo(AP4_UI32 pool_id, const
   fragment_pool_[pool_id].nal_length_size_ = nal_length_size;
   fragment_pool_[pool_id].annexb_sps_pps_.SetData(annexb_sps_pps.GetData(), annexb_sps_pps.GetDataSize());
   fragment_pool_[pool_id].decrypter_flags_ = flags;
+  fragment_pool_[pool_id].m_cryptoInfo = cryptoInfo;
 
   return AP4_SUCCESS;
 }
@@ -1093,9 +1103,9 @@ void WV_CencSingleSampleDecrypter::SetInput(cdm::InputBuffer_2& cdmInputBuffer,
   cdmInputBuffer.key_id = fragInfo.key_;
   cdmInputBuffer.key_id_size = 16;
   cdmInputBuffer.subsamples = subsamples.data();
-  cdmInputBuffer.encryption_scheme = media::ToCdmEncryptionScheme(m_EncryptionMode);
+  cdmInputBuffer.encryption_scheme = media::ToCdmEncryptionScheme(fragInfo.m_cryptoInfo.m_mode);
   cdmInputBuffer.timestamp = 0;
-  cdmInputBuffer.pattern = {m_CryptBlocks, m_SkipBlocks};
+  cdmInputBuffer.pattern = {fragInfo.m_cryptoInfo.m_cryptBlocks, fragInfo.m_cryptoInfo.m_skipBlocks};
 }
 
 /*----------------------------------------------------------------------
@@ -1256,43 +1266,9 @@ AP4_Result WV_CencSingleSampleDecrypter::DecryptSampleData(AP4_UI32 pool_id,
   cdm::Status ret{cdm::Status::kSuccess};
   std::vector<cdm::SubsampleEntry> subsamples;
   subsamples.reserve(subsample_count);
-  bool useSingleDecrypt{(fragInfo.decrypter_flags_ & SSD_DECRYPTER::SSD_CAPS::SSD_SINGLE_DECRYPT) !=
-                        0};
-  bool useCbcDecrypt{m_EncryptionMode == CryptoMode::AES_CBC};
+
+  bool useCbcDecrypt{fragInfo.m_cryptoInfo.m_mode == CryptoMode::AES_CBC};
   
-  // Decrypting (and not decoding) with subsamples > 1 seems to be broken for
-  // a long time now, we should consider removing support for this
-  if (!useSingleDecrypt) // The decrypter supports subsamples set > 1
-  {
-    uint32_t numCipherBytes{0};
-    for (size_t i{0}; i < subsample_count; i++)
-    {
-      subsamples.push_back({bytes_of_cleartext_data[i], bytes_of_encrypted_data[i]});
-      numCipherBytes += bytes_of_encrypted_data[i];
-    }
-
-    if (numCipherBytes == 0)
-    {
-      data_out.AppendData(data_in.GetData(), data_in.GetDataSize());
-      return AP4_SUCCESS;
-    }
-
-    cdm::InputBuffer_2 cdm_in;
-    SetInput(cdm_in, data_in, subsample_count, iv, fragInfo, subsamples);
-    data_out.SetDataSize(data_in.GetDataSize());
-    CdmBuffer buf{&data_out};
-    CdmDecryptedBlock cdm_out;
-    cdm_out.SetDecryptedBuffer(&buf);
-
-    CheckLicenseRenewal();
-    ret = drm_.GetCdmAdapter()->Decrypt(cdm_in, &cdm_out);
-    if (ret != cdm::Status::kSuccess)
-    {
-      LogDecryptError(ret, fragInfo.key_);
-    }
-    return (ret == cdm::Status::kSuccess) ? AP4_SUCCESS : AP4_ERROR_INVALID_PARAMETERS;
-  }
-
   // We can only decrypt with subsamples set to 1
   // This must be handled differently for CENC and CBCS
   // CENC:
@@ -1302,82 +1278,80 @@ AP4_Result WV_CencSingleSampleDecrypter::DecryptSampleData(AP4_UI32 pool_id,
   // from it before passing to CDM.
   // CBCS:
   // Due to the nature of this cipher subsamples must be decrypted separately
-  else
-  {
-    const unsigned int iterations{useCbcDecrypt ? subsample_count : 1};
-    size_t absPos{0};
 
-    for (unsigned int i{0}; i < iterations; ++i)
+  const unsigned int iterations{useCbcDecrypt ? subsample_count : 1};
+  size_t absPos{0};
+
+  for (unsigned int i{0}; i < iterations; ++i)
+  {
+    decrypt_in_.Reserve(data_in.GetDataSize());
+    decrypt_in_.SetDataSize(0);
+    size_t decryptInPos = absPos;
+    if (useCbcDecrypt)
+    {
+      UnpackSubsampleData(data_in, decryptInPos, i, bytes_of_cleartext_data,
+                          bytes_of_encrypted_data);
+    }
+    else
+    {
+      for (unsigned int subsamplePos{0}; subsamplePos < subsample_count; ++subsamplePos)
+      {
+        UnpackSubsampleData(data_in, absPos, subsamplePos, bytes_of_cleartext_data, bytes_of_encrypted_data);
+      }
+    }
+
+    if (decrypt_in_.GetDataSize() > 0) // remember to include when calling setcdmsubsamples
+    {
+      SetCdmSubsamples(subsamples, useCbcDecrypt);
+    }
+
+    else // we have nothing to decrypt in this iteration
     {
-      decrypt_in_.Reserve(data_in.GetDataSize());
-      decrypt_in_.SetDataSize(0);
-      size_t decryptInPos = absPos;
       if (useCbcDecrypt)
       {
-        UnpackSubsampleData(data_in, decryptInPos, i, bytes_of_cleartext_data,
+        data_out.AppendData(data_in.GetData() + absPos, bytes_of_cleartext_data[i]);
+        absPos += bytes_of_cleartext_data[i];
+        continue;
+      }
+      else // we can exit here for CENC and just return the input buffer
+      {
+        data_out.AppendData(data_in.GetData(), data_in.GetDataSize());
+        return AP4_SUCCESS;
+      }
+    }
+
+    cdm::InputBuffer_2 cdm_in;
+    SetInput(cdm_in, decrypt_in_, 1, iv, fragInfo, subsamples);
+    decrypt_out_.SetDataSize(decrypt_in_.GetDataSize());
+    CdmBuffer buf{&decrypt_out_};
+    CdmDecryptedBlock cdm_out;
+    cdm_out.SetDecryptedBuffer(&buf);
+
+    CheckLicenseRenewal();
+    ret = drm_.GetCdmAdapter()->Decrypt(cdm_in, &cdm_out);
+
+    if (ret == cdm::Status::kSuccess)
+    {
+      size_t cipherPos = 0;
+      if (useCbcDecrypt)
+      {
+        RepackSubsampleData(data_in, data_out, absPos, cipherPos, i, bytes_of_cleartext_data,
                             bytes_of_encrypted_data);
       }
       else
       {
-        for (unsigned int subsamplePos{0}; subsamplePos < subsample_count; ++subsamplePos)
-        {
-          UnpackSubsampleData(data_in, absPos, subsamplePos, bytes_of_cleartext_data, bytes_of_encrypted_data);
-        }
-      }
-
-      if (decrypt_in_.GetDataSize() > 0) // remember to include when calling setcdmsubsamples
-      {
-        SetCdmSubsamples(subsamples, useCbcDecrypt);
-      }
-
-      else // we have nothing to decrypt in this iteration
-      {
-        if (useCbcDecrypt)
-        {
-          data_out.AppendData(data_in.GetData() + absPos, bytes_of_cleartext_data[i]);
-          absPos += bytes_of_cleartext_data[i];
-          continue;
-        }
-        else // we can exit here for CENC and just return the input buffer
-        {
-          data_out.AppendData(data_in.GetData(), data_in.GetDataSize());
-          return AP4_SUCCESS;
-        }
-      }
-
-      cdm::InputBuffer_2 cdm_in;
-      SetInput(cdm_in, decrypt_in_, 1, iv, fragInfo, subsamples);
-      decrypt_out_.SetDataSize(decrypt_in_.GetDataSize());
-      CdmBuffer buf{&decrypt_out_};
-      CdmDecryptedBlock cdm_out;
-      cdm_out.SetDecryptedBuffer(&buf);
-
-      CheckLicenseRenewal();
-      ret = drm_.GetCdmAdapter()->Decrypt(cdm_in, &cdm_out);
-
-      if (ret == cdm::Status::kSuccess)
-      {
-        size_t cipherPos = 0;
-        if (useCbcDecrypt)
+        size_t absPos{0};
+        for (unsigned int i{0}; i < subsample_count; ++i)
         {
           RepackSubsampleData(data_in, data_out, absPos, cipherPos, i, bytes_of_cleartext_data,
                               bytes_of_encrypted_data);
         }
-        else
-        {
-          size_t absPos{0};
-          for (unsigned int i{0}; i < subsample_count; ++i)
-          {
-            RepackSubsampleData(data_in, data_out, absPos, cipherPos, i, bytes_of_cleartext_data,
-                                bytes_of_encrypted_data);
-          }
-        }
-      }
-      else
-      {
-        LogDecryptError(ret, fragInfo.key_);
       }
     }
+    else
+    {
+      LogDecryptError(ret, fragInfo.key_);
+    }
   }
   return (ret == cdm::Status::kSuccess) ? AP4_SUCCESS : AP4_ERROR_INVALID_PARAMETERS;
 }
@@ -1509,11 +1483,6 @@ void WV_CencSingleSampleDecrypter::ResetVideo()
   drained_ = true;
 }
 
-void WV_CencSingleSampleDecrypter::SetEncryptionMode(CryptoMode encryptionMode)
-{
-  m_EncryptionMode = encryptionMode;
-}
-
 void WV_CencSingleSampleDecrypter::SetDefaultKeyId(std::string_view keyId)
 {
   m_defaultKeyId = keyId;
@@ -1657,6 +1626,22 @@ private:
 
 extern "C" {
 
+// Linux arm64 version of libwidevinecdm.so depends on two
+// dynamic symbols. See https://github.com/xbmc/inputstream.adaptive/issues/1128
+#if defined(__linux__) && defined(__aarch64__) && !defined(ANDROID)
+__attribute__((target("no-outline-atomics")))
+int32_t __aarch64_ldadd4_acq_rel(int32_t value, int32_t *ptr)
+{
+  return __atomic_fetch_add(ptr, value, __ATOMIC_ACQ_REL);
+}
+
+__attribute__((target("no-outline-atomics")))
+int32_t __aarch64_swp4_acq_rel(int32_t value, int32_t *ptr)
+{
+  return __atomic_exchange_n(ptr, value, __ATOMIC_ACQ_REL);
+}
+#endif
+
 #ifdef _WIN32
 #define MODULE_API __declspec(dllexport)
 #else
